import pytest

import salt.loader
import salt.modules.win_lgpo as win_lgpo_module
import salt.states.win_lgpo as win_lgpo_state
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts, modules):
    loader = {
        "__opts__": minion_opts,
        "__salt__": modules,
        "__utils__": salt.loader.utils(minion_opts),
        "__context__": {},
    }
    return {
        win_lgpo_module: loader,
        win_lgpo_state: loader,
    }


@pytest.fixture
def stage_single():
    current_setting = win_lgpo_module.get_policy(
        policy_name="SeTakeOwnershipPrivilege",
        policy_class="machine",
    )
    try:
        win_lgpo_module.set_computer_policy(
            name="SeTakeOwnershipPrivilege",
            setting=["BUILTIN\\Administrators"],
            cumulative_rights_assignments=False,
        )
        yield
    finally:
        win_lgpo_module.set_computer_policy(
            name="SeTakeOwnershipPrivilege",
            setting=current_setting,
            cumulative_rights_assignments=False,
        )


@pytest.fixture
def stage_multiple():
    current_setting = win_lgpo_module.get_policy(
        policy_name="SeTakeOwnershipPrivilege",
        policy_class="machine",
    )
    try:
        win_lgpo_module.set_computer_policy(
            name="SeTakeOwnershipPrivilege",
            setting=["BUILTIN\\Administrators", "BUILTIN\\Backup Operators"],
            cumulative_rights_assignments=False,
        )
        yield
    finally:
        win_lgpo_module.set_computer_policy(
            name="SeTakeOwnershipPrivilege",
            setting=current_setting,
            cumulative_rights_assignments=False,
        )


def test_cumulative_rights_assignments(stage_single):
    expected = ["BUILTIN\\Administrators", "BUILTIN\\Backup Operators"]
    win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["BUILTIN\\Backup Operators"],
        policy_class="machine",
        cumulative_rights_assignments=True,
    )
    result = win_lgpo_module.get_policy(
        policy_name="SeTakeOwnershipPrivilege",
        policy_class="machine",
    )
    assert sorted(result) == sorted(expected)


def test_cumulative_rights_assignments_test_true(stage_single):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": None,
        "changes": {},
        "comment": (
            "The following policies are set to change:\nSeTakeOwnershipPrivilege"
        ),
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["BUILTIN\\Backup Operators"],
            policy_class="machine",
            cumulative_rights_assignments=True,
        )
    assert result == expected


def test_cumulative_rights_assignments_exists(stage_multiple):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    result = win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["BUILTIN\\Backup Operators"],
        policy_class="machine",
        cumulative_rights_assignments=True,
    )
    assert result == expected


def test_cumulative_rights_assignments_exists_test_true(stage_multiple):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["BUILTIN\\Backup Operators"],
            policy_class="machine",
            cumulative_rights_assignments=True,
        )
    assert result == expected


def test_non_cumulative_rights_assignments(stage_multiple):
    expected = ["BUILTIN\\Administrators"]
    win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["BUILTIN\\Administrators"],
        policy_class="machine",
        cumulative_rights_assignments=False,
    )
    result = win_lgpo_module.get_policy(
        policy_name="SeTakeOwnershipPrivilege",
        policy_class="machine",
    )
    assert sorted(result) == sorted(expected)


def test_non_cumulative_rights_assignments_test_true(stage_multiple):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": None,
        "changes": {},
        "comment": (
            "The following policies are set to change:\nSeTakeOwnershipPrivilege"
        ),
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["BUILTIN\\Administrators"],
            policy_class="machine",
            cumulative_rights_assignments=False,
        )
    assert result == expected


def test_non_cumulative_rights_assignments_exists(stage_single):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    result = win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["BUILTIN\\Administrators"],
        policy_class="machine",
        cumulative_rights_assignments=False,
    )
    assert result == expected


def test_non_cumulative_rights_assignments_exists_test_true(stage_single):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["BUILTIN\\Administrators"],
            policy_class="machine",
            cumulative_rights_assignments=False,
        )
    assert result == expected


def test_cumulative_rights_assignments_resolve_name(stage_single):
    expected = ["BUILTIN\\Administrators", "BUILTIN\\Backup Operators"]
    win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["Backup Operators"],
        policy_class="machine",
        cumulative_rights_assignments=True,
    )
    result = win_lgpo_module.get_policy(
        policy_name="SeTakeOwnershipPrivilege",
        policy_class="machine",
    )
    assert sorted(result) == sorted(expected)


def test_cumulative_rights_assignments_resolve_name_test_true(stage_single):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": None,
        "changes": {},
        "comment": (
            "The following policies are set to change:\nSeTakeOwnershipPrivilege"
        ),
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["Backup Operators"],
            policy_class="machine",
            cumulative_rights_assignments=True,
        )
    assert result == expected


def test_cumulative_rights_assignments_resolve_name_exists(stage_multiple):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    result = win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["Backup Operators"],
        policy_class="machine",
        cumulative_rights_assignments=True,
    )
    assert result == expected


def test_cumulative_rights_assignments_resolve_name_exists_test_true(stage_multiple):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["Backup Operators"],
            policy_class="machine",
            cumulative_rights_assignments=True,
        )
    assert result == expected


def test_non_cumulative_rights_assignments_resolve_name(stage_multiple):
    expected = ["BUILTIN\\Administrators"]
    win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["Administrators"],
        policy_class="machine",
        cumulative_rights_assignments=False,
    )
    result = win_lgpo_module.get_policy(
        policy_name="SeTakeOwnershipPrivilege",
        policy_class="machine",
    )
    assert sorted(result) == sorted(expected)


def test_non_cumulative_rights_assignments_resolve_name_test_true(stage_multiple):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": None,
        "changes": {},
        "comment": (
            "The following policies are set to change:\nSeTakeOwnershipPrivilege"
        ),
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["Administrators"],
            policy_class="machine",
            cumulative_rights_assignments=False,
        )
    assert result == expected


def test_non_cumulative_rights_assignments_resolve_name_exists(stage_single):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    result = win_lgpo_state.set_(
        name="SeTakeOwnershipPrivilege",
        setting=["Administrators"],
        policy_class="machine",
        cumulative_rights_assignments=False,
    )
    assert result == expected


def test_non_cumulative_rights_assignments_resolve_name_exists_test_true(stage_single):
    expected = {
        "name": "SeTakeOwnershipPrivilege",
        "result": True,
        "changes": {},
        "comment": "All specified policies are properly configured",
    }
    with patch.dict(win_lgpo_state.__opts__, {"test": True}):
        result = win_lgpo_state.set_(
            name="SeTakeOwnershipPrivilege",
            setting=["Administrators"],
            policy_class="machine",
            cumulative_rights_assignments=False,
        )
    assert result == expected



"""
Provide the service module for system supervisord or supervisord in a
virtualenv
"""


import configparser
import os

import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, CommandNotFoundError


def __virtual__():
    # We can't decide at load time whether supervisorctl is present. The
    # function _get_supervisorctl_bin does a much more thorough job and can
    # only be accurate at call time.
    return True


def _get_supervisorctl_bin(bin_env):
    """
    Return supervisorctl command to call, either from a virtualenv, an argument
    passed in, or from the global modules options
    """
    cmd = "supervisorctl"
    if not bin_env:
        which_result = __salt__["cmd.which_bin"]([cmd])
        if which_result is None:
            raise CommandNotFoundError("Could not find a `{}` binary".format(cmd))
        return which_result

    # try to get binary from env
    if os.path.isdir(bin_env):
        cmd_bin = os.path.join(bin_env, "bin", cmd)
        if os.path.isfile(cmd_bin):
            return cmd_bin
        raise CommandNotFoundError("Could not find a `{}` binary".format(cmd))

    return bin_env


def _ctl_cmd(cmd, name, conf_file, bin_env):
    """
    Return the command list to use
    """
    ret = [_get_supervisorctl_bin(bin_env)]
    if conf_file is not None:
        ret += ["-c", conf_file]
    ret.append(cmd)
    if name:
        ret.append(name)
    return ret


def _get_return(ret):
    retmsg = ret["stdout"]
    if ret["retcode"] != 0:
        # This is a non 0 exit code
        if "ERROR" not in retmsg:
            retmsg = "ERROR: {}".format(retmsg)
    return retmsg


def start(name="all", user=None, conf_file=None, bin_env=None):
    """
    Start the named service.
    Process group names should not include a trailing asterisk.

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.start <service>
        salt '*' supervisord.start <group>:
    """
    if name.endswith(":*"):
        name = name[:-1]
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("start", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def restart(name="all", user=None, conf_file=None, bin_env=None):
    """
    Restart the named service.
    Process group names should not include a trailing asterisk.

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.restart <service>
        salt '*' supervisord.restart <group>:
    """
    if name.endswith(":*"):
        name = name[:-1]
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("restart", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def stop(name="all", user=None, conf_file=None, bin_env=None):
    """
    Stop the named service.
    Process group names should not include a trailing asterisk.

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.stop <service>
        salt '*' supervisord.stop <group>:
    """
    if name.endswith(":*"):
        name = name[:-1]
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("stop", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def add(name, user=None, conf_file=None, bin_env=None):
    """
    Activates any updates in config for process/group.

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.add <name>
    """
    if name.endswith(":"):
        name = name[:-1]
    elif name.endswith(":*"):
        name = name[:-2]
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("add", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def remove(name, user=None, conf_file=None, bin_env=None):
    """
    Removes process/group from active config

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.remove <name>
    """
    if name.endswith(":"):
        name = name[:-1]
    elif name.endswith(":*"):
        name = name[:-2]
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("remove", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def reread(user=None, conf_file=None, bin_env=None):
    """
    Reload the daemon's configuration files

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.reread
    """
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("reread", None, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def update(user=None, conf_file=None, bin_env=None, name=None):
    """
    Reload config and add/remove/update as necessary

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed
    name
        name of the process group to update. if none then update any
        process group that has changes

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.update
    """

    if isinstance(name, str):
        if name.endswith(":"):
            name = name[:-1]
        elif name.endswith(":*"):
            name = name[:-2]

    ret = __salt__["cmd.run_all"](
        _ctl_cmd("update", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def status(name=None, user=None, conf_file=None, bin_env=None):
    """
    List programs and its state

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.status
    """
    all_process = {}
    for line in status_raw(name, user, conf_file, bin_env).splitlines():
        if len(line.split()) > 2:
            process, state, reason = line.split(None, 2)
        else:
            process, state, reason = line.split() + [""]
        all_process[process] = {"state": state, "reason": reason}
    return all_process


def status_raw(name=None, user=None, conf_file=None, bin_env=None):
    """
    Display the raw output of status

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.status_raw
    """
    ret = __salt__["cmd.run_all"](
        _ctl_cmd("status", name, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


def custom(command, user=None, conf_file=None, bin_env=None):
    """
    Run any custom supervisord command

    user
        user to run supervisorctl as
    conf_file
        path to supervisord config file
    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.custom "mstop '*gunicorn*'"
    """
    ret = __salt__["cmd.run_all"](
        _ctl_cmd(command, None, conf_file, bin_env),
        runas=user,
        python_shell=False,
    )
    return _get_return(ret)


# TODO: try to find a way to use the supervisor python module to read the
# config information
def _read_config(conf_file=None):
    """
    Reads the config file using configparser
    """
    if conf_file is None:
        paths = ("/etc/supervisor/supervisord.conf", "/etc/supervisord.conf")
        for path in paths:
            if os.path.exists(path):
                conf_file = path
                break
    if conf_file is None:
        raise CommandExecutionError("No suitable config file found")
    config = configparser.ConfigParser()
    try:
        config.read(conf_file)
    except OSError as exc:
        raise CommandExecutionError("Unable to read from {}: {}".format(conf_file, exc))
    return config


def options(name, conf_file=None):
    """
    .. versionadded:: 2014.1.0

    Read the config file and return the config options for a given process

    name
        Name of the configured process
    conf_file
        path to supervisord config file

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.options foo
    """
    config = _read_config(conf_file)
    section_name = "program:{}".format(name)
    if section_name not in config.sections():
        raise CommandExecutionError("Process '{}' not found".format(name))
    ret = {}
    for key, val in config.items(section_name):
        val = salt.utils.stringutils.to_num(val.split(";")[0].strip())
        # pylint: disable=maybe-no-member
        if isinstance(val, str):
            if val.lower() == "true":
                val = True
            elif val.lower() == "false":
                val = False
        # pylint: enable=maybe-no-member
        ret[key] = val
    return ret


def status_bool(name, expected_state=None, user=None, conf_file=None, bin_env=None):
    """
    Check for status of a specific supervisord process and return boolean result.

    name
        name of the process to check

    expected_state
        search for a specific process state. If set to ``None`` - any process state will match.

    user
        user to run supervisorctl as

    conf_file
        path to supervisord config file

    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    CLI Example:

    .. code-block:: bash

        salt '*' supervisord.status_bool nginx expected_state='RUNNING'
    """

    cmd = "status {}".format(name)
    for line in custom(cmd, user, conf_file, bin_env).splitlines():
        if len(line.split()) > 2:
            process, state, reason = line.split(None, 2)
        else:
            process, state, reason = line.split() + [""]

    if reason == "(no such process)" or process != name:
        return False

    if expected_state is None or state == expected_state:
        return True
    else:
        return False



"""
Return data to an influxdb server.

.. versionadded:: 2015.8.0

To enable this returner the minion will need the python client for influxdb
installed and the following values configured in the minion or master
config, these are the defaults:

.. code-block:: yaml

    influxdb.db: 'salt'
    influxdb.user: 'salt'
    influxdb.password: 'salt'
    influxdb.host: 'localhost'
    influxdb.port: 8086


Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location:

.. code-block:: yaml

    alternative.influxdb.db: 'salt'
    alternative.influxdb.user: 'salt'
    alternative.influxdb.password: 'salt'
    alternative.influxdb.host: 'localhost'
    alternative.influxdb.port: 6379

To use the influxdb returner, append '--return influxdb' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return influxdb

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return influxdb --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return influxdb --return_kwargs '{"db": "another-salt"}'

"""

import logging

import requests

import salt.returners
import salt.utils.jid
from salt.utils.decorators import memoize

try:
    import influxdb
    import influxdb.influxdb08

    HAS_INFLUXDB = True
except ImportError:
    HAS_INFLUXDB = False

# HTTP API header used to check the InfluxDB version
influxDBVersionHeader = "X-Influxdb-Version"

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "influxdb"


def __virtual__():
    if not HAS_INFLUXDB:
        return (
            False,
            "Could not import influxdb returner; "
            "influxdb python client is not installed.",
        )
    return __virtualname__


def _get_options(ret=None):
    """
    Get the influxdb options from salt.
    """
    attrs = {
        "host": "host",
        "port": "port",
        "db": "db",
        "user": "user",
        "password": "password",
    }

    _options = salt.returners.get_returner_options(
        __virtualname__, ret, attrs, __salt__=__salt__, __opts__=__opts__
    )
    return _options


@memoize
def _get_version(host, port, user, password):
    version = None
    # check the InfluxDB version via the HTTP API
    try:
        result = requests.get(
            "http://{}:{}/ping".format(host, port), auth=(user, password)
        )
        if influxDBVersionHeader in result.headers:
            version = result.headers[influxDBVersionHeader]
    except Exception as ex:  # pylint: disable=broad-except
        log.critical(
            "Failed to query InfluxDB version from HTTP API within InfluxDB "
            "returner: %s",
            ex,
        )
    return version


def _get_serv(ret=None):
    """
    Return an influxdb client object
    """
    _options = _get_options(ret)
    host = _options.get("host")
    port = _options.get("port")
    database = _options.get("db")
    user = _options.get("user")
    password = _options.get("password")
    version = _get_version(host, port, user, password)

    if version and "v0.8" in version:
        return influxdb.influxdb08.InfluxDBClient(
            host=host, port=port, username=user, password=password, database=database
        )
    else:
        return influxdb.InfluxDBClient(
            host=host, port=port, username=user, password=password, database=database
        )


def returner(ret):
    """
    Return data to a influxdb data store
    """
    serv = _get_serv(ret)

    # strip the 'return' key to avoid data duplication in the database
    json_return = salt.utils.json.dumps(ret["return"])
    del ret["return"]
    json_full_ret = salt.utils.json.dumps(ret)

    # create legacy request in case an InfluxDB 0.8.x version is used
    if "influxdb08" in serv.__module__:
        req = [
            {
                "name": "returns",
                "columns": ["fun", "id", "jid", "return", "full_ret"],
                "points": [
                    [ret["fun"], ret["id"], ret["jid"], json_return, json_full_ret]
                ],
            }
        ]
    # create InfluxDB 0.9+ version request
    else:
        req = [
            {
                "measurement": "returns",
                "tags": {"fun": ret["fun"], "id": ret["id"], "jid": ret["jid"]},
                "fields": {"return": json_return, "full_ret": json_full_ret},
            }
        ]

    try:
        serv.write_points(req)
    except Exception as ex:  # pylint: disable=broad-except
        log.critical("Failed to store return with InfluxDB returner: %s", ex)


def save_load(jid, load, minions=None):
    """
    Save the load to the specified jid
    """
    serv = _get_serv(ret=None)

    # create legacy request in case an InfluxDB 0.8.x version is used
    if "influxdb08" in serv.__module__:
        req = [
            {
                "name": "jids",
                "columns": ["jid", "load"],
                "points": [[jid, salt.utils.json.dumps(load)]],
            }
        ]
    # create InfluxDB 0.9+ version request
    else:
        req = [
            {
                "measurement": "jids",
                "tags": {"jid": jid},
                "fields": {"load": salt.utils.json.dumps(load)},
            }
        ]

    try:
        serv.write_points(req)
    except Exception as ex:  # pylint: disable=broad-except
        log.critical("Failed to store load with InfluxDB returner: %s", ex)


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    """
    Included for API consistency
    """


def get_load(jid):
    """
    Return the load data that marks a specified jid
    """
    serv = _get_serv(ret=None)
    sql = "select load from jids where jid = '{}'".format(jid)

    log.debug(">> Now in get_load %s", jid)
    data = serv.query(sql)
    log.debug(">> Now Data: %s", data)
    if data:
        return data
    return {}


def get_jid(jid):
    """
    Return the information returned when the specified job id was executed
    """
    serv = _get_serv(ret=None)

    sql = "select id, full_ret from returns where jid = '{}'".format(jid)

    data = serv.query(sql)
    ret = {}
    if data:
        points = data[0]["points"]
        for point in points:
            ret[point[3]] = salt.utils.json.loads(point[2])

    return ret


def get_fun(fun):
    """
    Return a dict of the last function called for all minions
    """
    serv = _get_serv(ret=None)

    sql = """select first(id) as fid, first(full_ret) as fret
            from returns
            where fun = '{}'
            group by fun, id
          """.format(
        fun
    )

    data = serv.query(sql)
    ret = {}
    if data:
        points = data[0]["points"]
        for point in points:
            ret[point[1]] = salt.utils.json.loads(point[2])

    return ret


def get_jids():
    """
    Return a list of all job ids
    """
    serv = _get_serv(ret=None)
    sql = "select distinct(jid) from jids group by load"

    # [{u'points': [[0, jid, load],
    #               [0, jid, load]],
    #   u'name': u'jids',
    #   u'columns': [u'time', u'distinct', u'load']}]
    data = serv.query(sql)
    ret = {}
    if data:
        for _, jid, load in data[0]["points"]:
            ret[jid] = salt.utils.jid.format_jid_instance(
                jid, salt.utils.json.loads(load)
            )
    return ret


def get_minions():
    """
    Return a list of minions
    """
    serv = _get_serv(ret=None)
    sql = "select distinct(id) from returns"

    data = serv.query(sql)
    ret = []
    if data:
        for jid in data[0]["points"]:
            ret.append(jid[1])

    return ret


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



"""
An engine that listens for libvirt events and resends them to the salt event bus.

The minimal configuration is the following and will listen to all events on the
local hypervisor and send them with a tag starting with ``salt/engines/libvirt_events``:

.. code-block:: yaml

    engines:
        - libvirt_events

Note that the automatically-picked libvirt connection will depend on the value
of ``uri_default`` in ``/etc/libvirt/libvirt.conf``. To force using another
connection like the local LXC libvirt driver, set the ``uri`` property as in the
following example configuration.

.. code-block:: yaml

    engines:
        - libvirt_events:
            uri: lxc:///
            tag_prefix: libvirt
            filters:
                - domain/lifecycle
                - domain/reboot
                - pool

Filters is a list of event types to relay to the event bus. Items in this list
can be either one of the main types (``domain``, ``network``, ``pool``,
``nodedev``, ``secret``), ``all`` or a more precise filter. These can be done
with values like <main_type>/<subtype>. The possible values are in the
CALLBACK_DEFS constant. If the filters list contains ``all``, all
events will be relayed.

Be aware that the list of events increases with libvirt versions, for example
network events have been added in libvirt 1.2.1 and storage events in 2.0.0.

Running the engine on non-root
------------------------------

Running this engine as non-root requires a special attention, which is surely
the case for the master running as user `salt`. The engine is likely to fail
to connect to libvirt with an error like this one:

    [ERROR   ] authentication unavailable: no polkit agent available to authenticate action 'org.libvirt.unix.monitor'


To fix this, the user running the engine, for example the salt-master, needs
to have the rights to connect to libvirt in the machine polkit config.
A polkit rule like the following one will allow `salt` user to connect to libvirt:

.. code-block:: javascript

    polkit.addRule(function(action, subject) {
        if (action.id.indexOf("org.libvirt") == 0 &&
            subject.user == "salt") {
            return polkit.Result.YES;
        }
    });

:depends: libvirt 1.0.0+ python binding

.. versionadded:: 2019.2.0
"""

import logging
import urllib.parse

import salt.utils.event

log = logging.getLogger(__name__)


try:
    import libvirt
except ImportError:
    libvirt = None  # pylint: disable=invalid-name


def __virtual__():
    """
    Only load if libvirt python binding is present
    """
    if libvirt is None:
        msg = "libvirt module not found"
    elif libvirt.getVersion() < 1000000:
        msg = "libvirt >= 1.0.0 required"
    else:
        msg = ""
    return not bool(msg), msg


REGISTER_FUNCTIONS = {
    "domain": "domainEventRegisterAny",
    "network": "networkEventRegisterAny",
    "pool": "storagePoolEventRegisterAny",
    "nodedev": "nodeDeviceEventRegisterAny",
    "secret": "secretEventRegisterAny",
}

# Handle either BLOCK_JOB or BLOCK_JOB_2, but prefer the latter
if hasattr(libvirt, "VIR_DOMAIN_EVENT_ID_BLOCK_JOB_2"):
    BLOCK_JOB_ID = "VIR_DOMAIN_EVENT_ID_BLOCK_JOB_2"
else:
    BLOCK_JOB_ID = "VIR_DOMAIN_EVENT_ID_BLOCK_JOB"

CALLBACK_DEFS = {
    "domain": (
        ("lifecycle", None),
        ("reboot", None),
        ("rtc_change", None),
        ("watchdog", None),
        ("graphics", None),
        ("io_error", "VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON"),
        ("control_error", None),
        ("disk_change", None),
        ("tray_change", None),
        ("pmwakeup", None),
        ("pmsuspend", None),
        ("balloon_change", None),
        ("pmsuspend_disk", None),
        ("device_removed", None),
        ("block_job", BLOCK_JOB_ID),
        ("tunable", None),
        ("agent_lifecycle", None),
        ("device_added", None),
        ("migration_iteration", None),
        ("job_completed", None),
        ("device_removal_failed", None),
        ("metadata_change", None),
        ("block_threshold", None),
    ),
    "network": (("lifecycle", None),),
    "pool": (
        ("lifecycle", "VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE"),
        ("refresh", "VIR_STORAGE_POOL_EVENT_ID_REFRESH"),
    ),
    "nodedev": (
        ("lifecycle", "VIR_NODE_DEVICE_EVENT_ID_LIFECYCLE"),
        ("update", "VIR_NODE_DEVICE_EVENT_ID_UPDATE"),
    ),
    "secret": (("lifecycle", None), ("value_changed", None)),
}


def _compute_subprefix(attr):
    """
    Get the part before the first '_' or the end of attr including
    the potential '_'
    """
    return "".join((attr.split("_")[0], "_" if len(attr.split("_")) > 1 else ""))


def _get_libvirt_enum_string(prefix, value):
    """
    Convert the libvirt enum integer value into a human readable string.

    :param prefix: start of the libvirt attribute to look for.
    :param value: integer to convert to string
    """
    attributes = [
        attr[len(prefix) :] for attr in libvirt.__dict__ if attr.startswith(prefix)
    ]

    # Filter out the values starting with a common base as they match another enum
    prefixes = [_compute_subprefix(p) for p in attributes]
    counts = {p: prefixes.count(p) for p in prefixes}
    sub_prefixes = [
        p
        for p, count in counts.items()
        if count > 1 or (p.endswith("_") and p[:-1] in prefixes)
    ]
    filtered = [
        attr for attr in attributes if _compute_subprefix(attr) not in sub_prefixes
    ]

    for candidate in filtered:
        if value == getattr(libvirt, "".join((prefix, candidate))):
            name = candidate.lower().replace("_", " ")
            return name
    return "unknown"


def _get_domain_event_detail(event, detail):
    """
    Convert event and detail numeric values into a tuple of human readable strings
    """
    event_name = _get_libvirt_enum_string("VIR_DOMAIN_EVENT_", event)
    if event_name == "unknown":
        return event_name, "unknown"

    prefix = "VIR_DOMAIN_EVENT_{}_".format(event_name.upper())
    detail_name = _get_libvirt_enum_string(prefix, detail)

    return event_name, detail_name


def _salt_send_event(opaque, conn, data):
    """
    Convenience function adding common data to the event and sending it
    on the salt event bus.

    :param opaque: the opaque data that is passed to the callback.
                   This is a dict with 'prefix', 'object' and 'event' keys.
    :param conn: libvirt connection
    :param data: additional event data dict to send
    """
    tag_prefix = opaque["prefix"]
    object_type = opaque["object"]
    event_type = opaque["event"]

    # Prepare the connection URI to fit in the tag
    # qemu+ssh://user@host:1234/system -> qemu+ssh/user@host:1234/system
    uri = urllib.parse.urlparse(conn.getURI())
    uri_tag = [uri.scheme]
    if uri.netloc:
        uri_tag.append(uri.netloc)
    path = uri.path.strip("/")
    if path:
        uri_tag.append(path)
    uri_str = "/".join(uri_tag)

    # Append some common data
    all_data = {"uri": conn.getURI()}
    all_data.update(data)

    tag = "/".join((tag_prefix, uri_str, object_type, event_type))

    # Actually send the event in salt
    if __opts__.get("__role") == "master":
        salt.utils.event.get_master_event(__opts__, __opts__["sock_dir"]).fire_event(
            all_data, tag
        )
    else:
        __salt__["event.send"](tag, all_data)


def _salt_send_domain_event(opaque, conn, domain, event, event_data):
    """
    Helper function send a salt event for a libvirt domain.

    :param opaque: the opaque data that is passed to the callback.
                   This is a dict with 'prefix', 'object' and 'event' keys.
    :param conn: libvirt connection
    :param domain: name of the domain related to the event
    :param event: name of the event
    :param event_data: additional event data dict to send
    """
    data = {
        "domain": {
            "name": domain.name(),
            "id": domain.ID(),
            "uuid": domain.UUIDString(),
        },
        "event": event,
    }
    data.update(event_data)
    _salt_send_event(opaque, conn, data)


def _domain_event_lifecycle_cb(conn, domain, event, detail, opaque):
    """
    Domain lifecycle events handler
    """
    event_str, detail_str = _get_domain_event_detail(event, detail)

    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {"event": event_str, "detail": detail_str},
    )


def _domain_event_reboot_cb(conn, domain, opaque):
    """
    Domain reboot events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {})


def _domain_event_rtc_change_cb(conn, domain, utcoffset, opaque):
    """
    Domain RTC change events handler
    """
    _salt_send_domain_event(
        opaque, conn, domain, opaque["event"], {"utcoffset": utcoffset}
    )


def _domain_event_watchdog_cb(conn, domain, action, opaque):
    """
    Domain watchdog events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {"action": _get_libvirt_enum_string("VIR_DOMAIN_EVENT_WATCHDOG_", action)},
    )


def _domain_event_io_error_cb(conn, domain, srcpath, devalias, action, reason, opaque):
    """
    Domain I/O Error events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "srcPath": srcpath,
            "dev": devalias,
            "action": _get_libvirt_enum_string("VIR_DOMAIN_EVENT_IO_ERROR_", action),
            "reason": reason,
        },
    )


def _domain_event_graphics_cb(
    conn, domain, phase, local, remote, auth, subject, opaque
):
    """
    Domain graphics events handler
    """
    prefix = "VIR_DOMAIN_EVENT_GRAPHICS_"

    def get_address(addr):
        """
        transform address structure into event data piece
        """
        return {
            "family": _get_libvirt_enum_string(
                "{}_ADDRESS_".format(prefix), addr["family"]
            ),
            "node": addr["node"],
            "service": addr["service"],
        }

    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "phase": _get_libvirt_enum_string(prefix, phase),
            "local": get_address(local),
            "remote": get_address(remote),
            "authScheme": auth,
            "subject": [{"type": item[0], "name": item[1]} for item in subject],
        },
    )


def _domain_event_control_error_cb(conn, domain, opaque):
    """
    Domain control error events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {})


def _domain_event_disk_change_cb(conn, domain, old_src, new_src, dev, reason, opaque):
    """
    Domain disk change events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "oldSrcPath": old_src,
            "newSrcPath": new_src,
            "dev": dev,
            "reason": _get_libvirt_enum_string("VIR_DOMAIN_EVENT_DISK_", reason),
        },
    )


def _domain_event_tray_change_cb(conn, domain, dev, reason, opaque):
    """
    Domain tray change events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "dev": dev,
            "reason": _get_libvirt_enum_string("VIR_DOMAIN_EVENT_TRAY_CHANGE_", reason),
        },
    )


def _domain_event_pmwakeup_cb(conn, domain, reason, opaque):
    """
    Domain wakeup events handler
    """
    _salt_send_domain_event(
        opaque, conn, domain, opaque["event"], {"reason": "unknown"}  # currently unused
    )


def _domain_event_pmsuspend_cb(conn, domain, reason, opaque):
    """
    Domain suspend events handler
    """
    _salt_send_domain_event(
        opaque, conn, domain, opaque["event"], {"reason": "unknown"}  # currently unused
    )


def _domain_event_balloon_change_cb(conn, domain, actual, opaque):
    """
    Domain balloon change events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {"actual": actual})


def _domain_event_pmsuspend_disk_cb(conn, domain, reason, opaque):
    """
    Domain disk suspend events handler
    """
    _salt_send_domain_event(
        opaque, conn, domain, opaque["event"], {"reason": "unknown"}  # currently unused
    )


def _domain_event_block_job_cb(conn, domain, disk, job_type, status, opaque):
    """
    Domain block job events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "disk": disk,
            "type": _get_libvirt_enum_string("VIR_DOMAIN_BLOCK_JOB_TYPE_", job_type),
            "status": _get_libvirt_enum_string("VIR_DOMAIN_BLOCK_JOB_", status),
        },
    )


def _domain_event_device_removed_cb(conn, domain, dev, opaque):
    """
    Domain device removal events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {"dev": dev})


def _domain_event_tunable_cb(conn, domain, params, opaque):
    """
    Domain tunable events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {"params": params})


# pylint: disable=invalid-name
def _domain_event_agent_lifecycle_cb(conn, domain, state, reason, opaque):
    """
    Domain agent lifecycle events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "state": _get_libvirt_enum_string(
                "VIR_CONNECT_DOMAIN_EVENT_AGENT_LIFECYCLE_STATE_", state
            ),
            "reason": _get_libvirt_enum_string(
                "VIR_CONNECT_DOMAIN_EVENT_AGENT_LIFECYCLE_REASON_", reason
            ),
        },
    )


def _domain_event_device_added_cb(conn, domain, dev, opaque):
    """
    Domain device addition events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {"dev": dev})


# pylint: disable=invalid-name
def _domain_event_migration_iteration_cb(conn, domain, iteration, opaque):
    """
    Domain migration iteration events handler
    """
    _salt_send_domain_event(
        opaque, conn, domain, opaque["event"], {"iteration": iteration}
    )


def _domain_event_job_completed_cb(conn, domain, params, opaque):
    """
    Domain job completion events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {"params": params})


def _domain_event_device_removal_failed_cb(conn, domain, dev, opaque):
    """
    Domain device removal failure events handler
    """
    _salt_send_domain_event(opaque, conn, domain, opaque["event"], {"dev": dev})


def _domain_event_metadata_change_cb(conn, domain, mtype, nsuri, opaque):
    """
    Domain metadata change events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {
            "type": _get_libvirt_enum_string("VIR_DOMAIN_METADATA_", mtype),
            "nsuri": nsuri,
        },
    )


def _domain_event_block_threshold_cb(
    conn, domain, dev, path, threshold, excess, opaque
):
    """
    Domain block threshold events handler
    """
    _salt_send_domain_event(
        opaque,
        conn,
        domain,
        opaque["event"],
        {"dev": dev, "path": path, "threshold": threshold, "excess": excess},
    )


def _network_event_lifecycle_cb(conn, net, event, detail, opaque):
    """
    Network lifecycle events handler
    """

    _salt_send_event(
        opaque,
        conn,
        {
            "network": {"name": net.name(), "uuid": net.UUIDString()},
            "event": _get_libvirt_enum_string("VIR_NETWORK_EVENT_", event),
            "detail": "unknown",  # currently unused
        },
    )


def _pool_event_lifecycle_cb(conn, pool, event, detail, opaque):
    """
    Storage pool lifecycle events handler
    """
    _salt_send_event(
        opaque,
        conn,
        {
            "pool": {"name": pool.name(), "uuid": pool.UUIDString()},
            "event": _get_libvirt_enum_string("VIR_STORAGE_POOL_EVENT_", event),
            "detail": "unknown",  # currently unused
        },
    )


def _pool_event_refresh_cb(conn, pool, opaque):
    """
    Storage pool refresh events handler
    """
    _salt_send_event(
        opaque,
        conn,
        {
            "pool": {"name": pool.name(), "uuid": pool.UUIDString()},
            "event": opaque["event"],
        },
    )


def _nodedev_event_lifecycle_cb(conn, dev, event, detail, opaque):
    """
    Node device lifecycle events handler
    """
    _salt_send_event(
        opaque,
        conn,
        {
            "nodedev": {"name": dev.name()},
            "event": _get_libvirt_enum_string("VIR_NODE_DEVICE_EVENT_", event),
            "detail": "unknown",  # currently unused
        },
    )


def _nodedev_event_update_cb(conn, dev, opaque):
    """
    Node device update events handler
    """
    _salt_send_event(
        opaque, conn, {"nodedev": {"name": dev.name()}, "event": opaque["event"]}
    )


def _secret_event_lifecycle_cb(conn, secret, event, detail, opaque):
    """
    Secret lifecycle events handler
    """
    _salt_send_event(
        opaque,
        conn,
        {
            "secret": {"uuid": secret.UUIDString()},
            "event": _get_libvirt_enum_string("VIR_SECRET_EVENT_", event),
            "detail": "unknown",  # currently unused
        },
    )


def _secret_event_value_changed_cb(conn, secret, opaque):
    """
    Secret value change events handler
    """
    _salt_send_event(
        opaque,
        conn,
        {"secret": {"uuid": secret.UUIDString()}, "event": opaque["event"]},
    )


def _cleanup(cnx):
    """
    Close the libvirt connection

    :param cnx: libvirt connection
    """
    log.debug("Closing libvirt connection: %s", cnx.getURI())
    cnx.close()


def _callbacks_cleanup(cnx, callback_ids):
    """
    Unregister all the registered callbacks

    :param cnx: libvirt connection
    :param callback_ids: dictionary mapping a libvirt object type to an ID list
                         of callbacks to deregister
    """
    for obj, ids in callback_ids.items():
        register_name = REGISTER_FUNCTIONS[obj]
        deregister_name = register_name.replace("Reg", "Dereg")
        deregister = getattr(cnx, deregister_name)
        for callback_id in ids:
            deregister(callback_id)


def _register_callback(cnx, tag_prefix, obj, event, real_id):
    """
    Helper function registering a callback

    :param cnx: libvirt connection
    :param tag_prefix: salt event tag prefix to use
    :param obj: the libvirt object name for the event. Needs to
                be one of the REGISTER_FUNCTIONS keys.
    :param event: the event type name.
    :param real_id: the libvirt name of an alternative event id to use or None

    :rtype integer value needed to deregister the callback
    """
    libvirt_name = real_id
    if real_id is None:
        libvirt_name = "VIR_{}_EVENT_ID_{}".format(obj, event).upper()

    if not hasattr(libvirt, libvirt_name):
        log.warning('Skipping "%s/%s" events: libvirt too old', obj, event)
        return None

    libvirt_id = getattr(libvirt, libvirt_name)
    callback_name = "_{}_event_{}_cb".format(obj, event)
    callback = globals().get(callback_name, None)
    if callback is None:
        log.error("Missing function %s in engine", callback_name)
        return None

    register = getattr(cnx, REGISTER_FUNCTIONS[obj])
    return register(
        None,
        libvirt_id,
        callback,
        {"prefix": tag_prefix, "object": obj, "event": event},
    )


def _append_callback_id(ids, obj, callback_id):
    """
    Helper function adding a callback ID to the IDs dict.
    The callback ids dict maps an object to event callback ids.

    :param ids: dict of callback IDs to update
    :param obj: one of the keys of REGISTER_FUNCTIONS
    :param callback_id: the result of _register_callback
    """
    if obj not in ids:
        ids[obj] = []
    ids[obj].append(callback_id)


def start(uri=None, tag_prefix="salt/engines/libvirt_events", filters=None):
    """
    Listen to libvirt events and forward them to salt.

    :param uri: libvirt URI to listen on.
                Defaults to None to pick the first available local hypervisor
    :param tag_prefix: the beginning of the salt event tag to use.
                       Defaults to 'salt/engines/libvirt_events'
    :param filters: the list of event of listen on. Defaults to 'all'
    """
    if filters is None:
        filters = ["all"]
    try:
        libvirt.virEventRegisterDefaultImpl()

        cnx = libvirt.openReadOnly(uri)
        log.debug("Opened libvirt uri: %s", cnx.getURI())

        callback_ids = {}
        all_filters = "all" in filters

        for obj, event_defs in CALLBACK_DEFS.items():
            for event, real_id in event_defs:
                event_filter = "/".join((obj, event))
                if (
                    event_filter not in filters
                    and obj not in filters
                    and not all_filters
                ):
                    continue
                registered_id = _register_callback(cnx, tag_prefix, obj, event, real_id)
                if registered_id:
                    _append_callback_id(callback_ids, obj, registered_id)

        exit_loop = False
        while not exit_loop:
            exit_loop = libvirt.virEventRunDefaultImpl() < 0

    except Exception as err:  # pylint: disable=broad-except
        log.exception(err)
    finally:
        _callbacks_cleanup(cnx, callback_ids)
        _cleanup(cnx)



"""
Management of NFS exports
===============================================

.. versionadded:: 2018.3.0

To ensure an NFS export exists:

.. code-block:: yaml

    add_simple_export:
      nfs_export.present:
        - name:     '/srv/nfs'
        - hosts:    '10.0.2.0/24'
        - options:
          - 'rw'

This creates the following in /etc/exports:

.. code-block:: bash

    /srv/nfs 10.0.2.0/24(rw)

For more complex exports with multiple groups of hosts, use 'clients':

.. code-block:: yaml

    add_complex_export:
      nfs_export.present:
        - name: '/srv/nfs'
        - clients:
          # First export, same as simple one above
          - hosts: '10.0.2.0/24'
            options:
              - 'rw'
          # Second export
          - hosts: '*.example.com'
            options:
              - 'ro'
              - 'subtree_check'

This creates the following in /etc/exports:

.. code-block:: bash

    /srv/nfs 10.0.2.0/24(rw) 192.168.0.0/24,172.19.0.0/16(ro,subtree_check)

Any export of the given path will be modified to match the one specified.

To ensure an NFS export is absent:

.. code-block:: yaml

    delete_export:
      nfs_export.absent:
        - name: '/srv/nfs'

"""

import salt.utils.path


def __virtual__():
    """
    Only work with nfs tools installed
    """
    cmd = "exportfs"
    if salt.utils.path.which(cmd):
        return bool(cmd)

    return (
        False,
        "The nfs_exports state module failed to load: "
        "the exportfs binary is not in the path",
    )


def present(name, clients=None, hosts=None, options=None, exports="/etc/exports"):
    """
    Ensure that the named export is present with the given options

    name
        The export path to configure

    clients
        A list of hosts and the options applied to them.
        This option may not be used in combination with
        the 'hosts' or 'options' shortcuts.

    .. code-block:: yaml

        - clients:
          # First export
          - hosts: '10.0.2.0/24'
            options:
              - 'rw'
          # Second export
          - hosts: '*.example.com'
            options:
              - 'ro'
              - 'subtree_check'

    hosts
        A string matching a number of hosts, for example:

    .. code-block:: yaml

        hosts: '10.0.2.123'

        hosts: '10.0.2.0/24'

        hosts: 'minion1.example.com'

        hosts: '*.example.com'

        hosts: '*'

    options
        A list of NFS options, for example:

    .. code-block:: yaml

        options:
          - 'rw'
          - 'subtree_check'

    """
    path = name
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if not clients:
        if not hosts:
            ret["result"] = False
            ret["comment"] = "Either 'clients' or 'hosts' must be defined"
            return ret
        # options being None is handled by add_export()
        clients = [{"hosts": hosts, "options": options}]

    old = __salt__["nfs3.list_exports"](exports)
    if path in old:
        if old[path] == clients:
            ret["result"] = True
            ret["comment"] = "Export {} already configured".format(path)
            return ret

        ret["changes"]["new"] = clients
        ret["changes"]["old"] = old[path]
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Export {} would be changed".format(path)
            return ret

        __salt__["nfs3.del_export"](exports, path)

    else:
        ret["changes"]["old"] = None
        ret["changes"]["new"] = clients
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Export {} would be added".format(path)
            return ret

    add_export = __salt__["nfs3.add_export"]
    for exp in clients:
        add_export(exports, path, exp["hosts"], exp["options"])

    ret["changes"]["new"] = clients

    try_reload = __salt__["nfs3.reload_exports"]()
    ret["comment"] = try_reload["stderr"]
    ret["result"] = try_reload["result"]
    return ret


def absent(name, exports="/etc/exports"):
    """
    Ensure that the named path is not exported

    name
        The export path to remove
    """

    path = name
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    old = __salt__["nfs3.list_exports"](exports)
    if path in old:
        if __opts__["test"]:
            ret["comment"] = "Export {} would be removed".format(path)
            ret["changes"][path] = old[path]
            ret["result"] = None
            return ret

        __salt__["nfs3.del_export"](exports, path)
        try_reload = __salt__["nfs3.reload_exports"]()
        if not try_reload["result"]:
            ret["comment"] = try_reload["stderr"]
        else:
            ret["comment"] = "Export {} removed".format(path)

        ret["result"] = try_reload["result"]
        ret["changes"][path] = old[path]
    else:
        ret["comment"] = "Export {} already absent".format(path)
        ret["result"] = True

    return ret



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.pw_group as pw_group
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pw_group: {"grinfo": {}}}


def test_add():
    """
    Tests to add the specified group
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(pw_group.__salt__, {"cmd.run_all": mock}):
        assert pw_group.add("a")


def test_delete():
    """
    Tests to remove the named group
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(pw_group.__salt__, {"cmd.run_all": mock}):
        assert pw_group.delete("a")


@pytest.mark.skip_on_windows(reason="grp not available on Windows")
def test_info():
    """
    Tests to return information about a group
    """
    assert pw_group.info("name") == {}

    mock = MagicMock(
        return_value={
            "gr_name": "A",
            "gr_passwd": "B",
            "gr_gid": 1,
            "gr_mem": ["C", "D"],
        }
    )
    with patch.dict(pw_group.grinfo, mock):
        assert pw_group.info("name") == {}


@pytest.mark.skip_on_windows(reason="grp not available on Windows")
def test_getent():
    """
    Tests for return info on all groups
    """
    mock_getent = [{"passwd": "x", "gid": 0, "name": "root"}]
    with patch.dict(pw_group.__context__, {"group.getent": mock_getent}):
        assert {"passwd": "x", "gid": 0, "name": "root"} == pw_group.getent()[0]

    mock = MagicMock(return_value="A")
    with patch.object(pw_group, "info", mock):
        assert pw_group.getent(True)[0] == "A"


def test_chgid():
    """
    tests to change the gid for a named group
    """
    mock = MagicMock(return_value=1)
    with patch.dict(pw_group.__salt__, {"file.group_to_gid": mock}):
        assert pw_group.chgid("name", 1)

    mock = MagicMock(side_effect=[1, 0])
    with patch.dict(pw_group.__salt__, {"file.group_to_gid": mock}):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_group.__salt__, {"cmd.run": mock}):
            assert pw_group.chgid("name", 0)

    mock = MagicMock(side_effect=[1, 1])
    with patch.dict(pw_group.__salt__, {"file.group_to_gid": mock}):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_group.__salt__, {"cmd.run": mock}):
            assert not pw_group.chgid("name", 0)



"""
unit tests for the fileserver runner
"""


import pytest

import salt.loader
import salt.runners.fileserver as fileserver
import salt.utils.files
from tests.support.mock import MagicMock, patch


class DummyFS:
    """
    Dummy object to provide the attributes needed to run unit tests
    """

    def __init__(self, backends):
        self.backends = backends

    def keys(self):
        return ["{}.envs".format(x) for x in self.backends]


@pytest.fixture
def cachedir(tmp_path):
    return tmp_path / "cache"


@pytest.fixture
def configure_loader_modules():
    return {fileserver: {"__opts__": {"extension_modules": ""}}}


def _make_file_lists_cache(cachedir, backends):
    """
    Create some dummy files to represent file list caches, as well as other
    files that aren't file list caches, so that we can confirm that *only*
    the cache files are touched. Create a dir for each configured backend,
    as well as for the roots backend (which is *not* configured as a
    backend in this test), so that we can ensure that its cache is left
    alone.
    """
    for back in backends:
        back_cachedir = cachedir / "file_lists" / back
        back_cachedir.mkdir(parents=True, exist_ok=True)
        for filename in ("base.p", "dev.p", "foo.txt"):
            (back_cachedir / filename).touch()


def test_clear_file_list_cache_vcs(cachedir):
    """
    Test that VCS backends are cleared irrespective of whether they are
    configured as gitfs/git, hgfs/hg, svnfs/svn.
    """
    # Mixture of VCS backends specified with and without "fs" at the end,
    # to confirm that the correct dirs are cleared.
    backends = ["gitfs", "hg", "svnfs"]
    opts = {
        "fileserver_backend": backends,
        "cachedir": str(cachedir),
    }
    mock_fs = DummyFS(backends)

    _make_file_lists_cache(cachedir, backends + ["roots"])

    with patch.dict(fileserver.__opts__, opts), patch.object(
        salt.loader, "fileserver", MagicMock(return_value=mock_fs)
    ):
        cleared = fileserver.clear_file_list_cache()

    # Make sure the return data matches what you'd expect
    expected = {
        "gitfs": ["base", "dev"],
        "hg": ["base", "dev"],
        "svnfs": ["base", "dev"],
    }
    assert cleared == expected, cleared

    # Trust, but verify! Check that the correct files are actually gone
    assert not (cachedir / "file_lists" / "gitfs" / "base.p").exists()
    assert not (cachedir / "file_lists" / "gitfs" / "dev.p").exists()
    assert not (cachedir / "file_lists" / "hg" / "base.p").exists()
    assert not (cachedir / "file_lists" / "gitfs" / "dev.p").exists()
    assert not (cachedir / "file_lists" / "hg" / "base.p").exists()
    assert not (cachedir / "file_lists" / "svnfs" / "dev.p").exists()

    # These files *should* exist and shouldn't have been cleaned
    assert (cachedir / "file_lists" / "gitfs" / "foo.txt").exists()
    assert (cachedir / "file_lists" / "hg" / "foo.txt").exists()
    assert (cachedir / "file_lists" / "svnfs" / "foo.txt").exists()
    assert (cachedir / "file_lists" / "roots" / "base.p").exists()
    assert (cachedir / "file_lists" / "roots" / "dev.p").exists()
    assert (cachedir / "file_lists" / "roots" / "foo.txt").exists()


def test_clear_file_list_cache_vcs_limited(cachedir):
    """
    Test the arguments to limit what is cleared
    """
    # Mixture of VCS backends specified with and without "fs" at the end,
    # to confirm that the correct dirs are cleared.
    backends = ["gitfs", "hg", "svnfs"]
    opts = {
        "fileserver_backend": backends,
        "cachedir": str(cachedir),
    }
    mock_fs = DummyFS(backends)

    _make_file_lists_cache(cachedir, backends + ["roots"])

    with patch.dict(fileserver.__opts__, opts), patch.object(
        salt.loader, "fileserver", MagicMock(return_value=mock_fs)
    ):
        cleared = fileserver.clear_file_list_cache(saltenv="base", backend="gitfs")

    expected = {"gitfs": ["base"]}
    assert cleared == expected, cleared

    # Trust, but verify! Check that the correct files are actually gone
    assert not (cachedir / "file_lists" / "gitfs" / "base.p").exists()

    # These files *should* exist and shouldn't have been cleaned
    assert (cachedir / "file_lists" / "gitfs" / "dev.p").exists()
    assert (cachedir / "file_lists" / "gitfs" / "foo.txt").exists()
    assert (cachedir / "file_lists" / "hg" / "base.p").exists()
    assert (cachedir / "file_lists" / "hg" / "dev.p").exists()
    assert (cachedir / "file_lists" / "hg" / "foo.txt").exists()
    assert (cachedir / "file_lists" / "svnfs" / "base.p").exists()
    assert (cachedir / "file_lists" / "svnfs" / "dev.p").exists()
    assert (cachedir / "file_lists" / "svnfs" / "foo.txt").exists()
    assert (cachedir / "file_lists" / "roots" / "base.p").exists()
    assert (cachedir / "file_lists" / "roots" / "dev.p").exists()
    assert (cachedir / "file_lists" / "roots" / "foo.txt").exists()



import pytest
from salt.utils.templates import render_genshi_tmpl

pytest.importorskip("genshi")


def test_render_sanity(render_context):
    tmpl = """<RU>OK</RU>"""
    res = render_genshi_tmpl(tmpl, render_context)
    assert res == "<RU>OK</RU>"


def test_render_evaluate(render_context):
    tmpl = """<RU>${ "OK" }</RU>"""
    res = render_genshi_tmpl(tmpl, render_context)
    assert res == "<RU>OK</RU>"


def test_render_evaluate_condition(render_context):
    tmpl = """<RU xmlns:py="http://genshi.edgewall.org/" py:if="1">OK</RU>"""
    res = render_genshi_tmpl(tmpl, render_context)
    assert res == "<RU>OK</RU>"


def test_render_variable(render_context):
    tmpl = """<RU>$var</RU>"""
    render_context["var"] = "OK"
    res = render_genshi_tmpl(tmpl, render_context)
    assert res == "<RU>OK</RU>"


def test_render_variable_replace(render_context):
    tmpl = """<RU xmlns:py="http://genshi.edgewall.org/" py:content="var">not ok</RU>"""
    render_context["var"] = "OK"
    res = render_genshi_tmpl(tmpl, render_context)
    assert res == "<RU>OK</RU>"



"""
Edit ini files

:maintainer: <akilesh1597@gmail.com>
:maturity: new
:depends: re
:platform: all

(for example /etc/sysctl.conf)
"""

import logging
import os
import re

import salt.utils.data
import salt.utils.files
import salt.utils.json
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

__virtualname__ = "ini"


def __virtual__():
    """
    Rename to ini
    """
    return __virtualname__


INI_REGX = re.compile(r"^\s*\[(.+?)\]\s*$", flags=re.M)
COM_REGX = re.compile(r"^\s*(#|;)\s*(.*)")
INDENTED_REGX = re.compile(r"(\s+)(.*)")


def set_option(file_name, sections=None, separator="="):
    """
    Edit an ini file, replacing one or more sections. Returns a dictionary
    containing the changes made.

    file_name
        path of ini_file

    sections : None
        A dictionary representing the sections to be edited ini file
        The keys are the section names and the values are the dictionary
        containing the options
        If the ini file does not contain sections the keys and values represent
        the options

    separator : =
        A character used to separate keys and values. Standard ini files use
        the "=" character.

        .. versionadded:: 2016.11.0

    API Example:

    .. code-block:: python

        import salt.client
        with salt.client.get_local_client() as sc:
            sc.cmd(
                'target', 'ini.set_option',
                ['path_to_ini_file', '{"section_to_change": {"key": "value"}}']
            )

    CLI Example:

    .. code-block:: bash

        salt '*' ini.set_option /path/to/ini '{section_foo: {key: value}}'
    """
    sections = sections or {}
    changes = {}
    inifile = _Ini.get_ini_file(file_name, separator=separator)
    changes = inifile.update(sections)
    inifile.flush()
    return changes


def get_option(file_name, section, option, separator="="):
    """
    Get value of a key from a section in an ini file. Returns ``None`` if
    no matching key was found.

    API Example:

    .. code-block:: python

        import salt.client
        with salt.client.get_local_client() as sc:
            sc.cmd('target', 'ini.get_option',
                   [path_to_ini_file, section_name, option])

    CLI Example:

    .. code-block:: bash

        salt '*' ini.get_option /path/to/ini section_name option_name
    """
    inifile = _Ini.get_ini_file(file_name, separator=separator)
    if section:
        try:
            return inifile.get(section, {}).get(option, None)
        except AttributeError:
            return None
    else:
        return inifile.get(option, None)


def remove_option(file_name, section, option, separator="="):
    """
    Remove a key/value pair from a section in an ini file. Returns the value of
    the removed key, or ``None`` if nothing was removed.

    API Example:

    .. code-block:: python

        import salt
        sc = salt.client.get_local_client()
        sc.cmd('target', 'ini.remove_option',
               [path_to_ini_file, section_name, option])

    CLI Example:

    .. code-block:: bash

        salt '*' ini.remove_option /path/to/ini section_name option_name
    """
    inifile = _Ini.get_ini_file(file_name, separator=separator)
    if isinstance(inifile.get(section), (dict, OrderedDict)):
        value = inifile.get(section, {}).pop(option, None)
    else:
        value = inifile.pop(option, None)
    inifile.flush()
    return value


def get_section(file_name, section, separator="="):
    """
    Retrieve a section from an ini file. Returns the section as dictionary. If
    the section is not found, an empty dictionary is returned.

    API Example:

    .. code-block:: python

        import salt.client
        with salt.client.get_local_client() as sc:
            sc.cmd('target', 'ini.get_section',
                   [path_to_ini_file, section_name])

    CLI Example:

    .. code-block:: bash

        salt '*' ini.get_section /path/to/ini section_name
    """
    inifile = _Ini.get_ini_file(file_name, separator=separator)
    ret = {}
    for key, value in inifile.get(section, {}).items():
        if key[0] != "#":
            ret.update({key: value})
    return ret


def remove_section(file_name, section, separator="="):
    """
    Remove a section in an ini file. Returns the removed section as dictionary,
    or ``None`` if nothing was removed.

    API Example:

    .. code-block:: python

        import salt.client
        with  salt.client.get_local_client() as sc:
            sc.cmd('target', 'ini.remove_section',
                   [path_to_ini_file, section_name])

    CLI Example:

    .. code-block:: bash

        salt '*' ini.remove_section /path/to/ini section_name
    """
    inifile = _Ini.get_ini_file(file_name, separator=separator)
    if section in inifile:
        section = inifile.pop(section)
        inifile.flush()
        ret = {}
        for key, value in section.items():
            if key[0] != "#":
                ret.update({key: value})
        return ret


def get_ini(file_name, separator="="):
    """
    Retrieve whole structure from an ini file and return it as dictionary.

    API Example:

    .. code-block:: python

        import salt.client
        with salt.client.giet_local_client() as sc:
            sc.cmd('target', 'ini.get_ini',
                   [path_to_ini_file])

    CLI Example:

    .. code-block:: bash

        salt '*' ini.get_ini /path/to/ini
    """

    def ini_odict2dict(odict):
        """
        Transform OrderedDict to regular dict recursively
        :param odict: OrderedDict
        :return: regular dict
        """
        ret = {}
        for key, val in odict.items():
            if key[0] != "#":
                if isinstance(val, (dict, OrderedDict)):
                    ret.update({key: ini_odict2dict(val)})
                else:
                    ret.update({key: val})
        return ret

    inifile = _Ini.get_ini_file(file_name, separator=separator)
    return ini_odict2dict(inifile)


class _Section(OrderedDict):
    def __init__(self, name, inicontents="", separator="=", commenter="#"):
        super().__init__(self)
        self.name = name
        self.inicontents = inicontents
        self.sep = separator
        self.com = commenter

        opt_regx_prefix = r"(\s*)(.+?)\s*"
        opt_regx_suffix = r"\s*(.*)\s*"
        self.opt_regx_str = r"{}(\{}){}".format(
            opt_regx_prefix, self.sep, opt_regx_suffix
        )
        self.opt_regx = re.compile(self.opt_regx_str)

    def refresh(self, inicontents=None):
        comment_count = 1
        unknown_count = 1
        curr_indent = ""
        inicontents = inicontents or self.inicontents
        inicontents = inicontents.strip(os.linesep)

        if not inicontents:
            return
        for opt in self:
            self.pop(opt)
        for opt_str in inicontents.split(os.linesep):
            # Match comments
            com_match = COM_REGX.match(opt_str)
            if com_match:
                name = "#comment{}".format(comment_count)
                self.com = com_match.group(1)
                comment_count += 1
                self.update({name: opt_str})
                continue
            # Add indented lines to the value of the previous entry.
            indented_match = INDENTED_REGX.match(opt_str)
            if indented_match:
                indent = indented_match.group(1).replace("\t", "    ")
                if indent > curr_indent:
                    options = list(self)
                    if options:
                        prev_opt = options[-1]
                        value = self.get(prev_opt)
                        self.update({prev_opt: os.linesep.join((value, opt_str))})
                    continue
            # Match normal key+value lines.
            opt_match = self.opt_regx.match(opt_str)
            if opt_match:
                curr_indent, name, self.sep, value = opt_match.groups()
                curr_indent = curr_indent.replace("\t", "    ")
                self.update({name: value})
                continue
            # Anything remaining is a mystery.
            name = "#unknown{}".format(unknown_count)
            self.update({name: opt_str})
            unknown_count += 1

    def _uncomment_if_commented(self, opt_key):
        # should be called only if opt_key is not already present
        # will uncomment the key if commented and create a place holder
        # for the key where the correct value can be update later
        # used to preserve the ordering of comments and commented options
        # and to make sure options without sectons go above any section
        options_backup = OrderedDict()
        comment_index = None
        for key, value in self.items():
            if comment_index is not None:
                options_backup.update({key: value})
                continue
            if "#comment" not in key:
                continue
            opt_match = self.opt_regx.match(value.lstrip("#"))
            if opt_match and opt_match.group(2) == opt_key:
                comment_index = key
        for key in options_backup:
            self.pop(key)
        self.pop(comment_index, None)
        super().update({opt_key: None})
        for key, value in options_backup.items():
            super().update({key: value})

    def update(self, update_dict):
        changes = {}
        for key, value in update_dict.items():
            # Ensure the value is either a _Section or a string
            if isinstance(value, (dict, OrderedDict)):
                sect = _Section(
                    name=key, inicontents="", separator=self.sep, commenter=self.com
                )
                sect.update(value)
                value = sect
                value_plain = value.as_dict()
            else:
                value = str(value)
                value_plain = value

            if key not in self:
                changes.update({key: {"before": None, "after": value_plain}})
                # If it's not a section, it may already exist as a
                # commented-out key/value pair
                if not isinstance(value, _Section):
                    self._uncomment_if_commented(key)

                super().update({key: value})
            else:
                curr_value = self.get(key, None)
                if isinstance(curr_value, _Section):
                    sub_changes = curr_value.update(value)
                    if sub_changes:
                        changes.update({key: sub_changes})
                else:
                    if curr_value != value:
                        changes.update(
                            {key: {"before": curr_value, "after": value_plain}}
                        )
                        super().update({key: value})
        return changes

    def gen_ini(self):
        yield "{0}[{1}]{0}".format(os.linesep, self.name)
        sections_dict = OrderedDict()
        for name, value in self.items():
            # Handle Comment Lines
            if COM_REGX.match(name):
                yield "{}{}".format(value, os.linesep)
            # Handle Sections
            elif isinstance(value, _Section):
                sections_dict.update({name: value})
            # Key / Value pairs
            # Adds spaces between the separator
            else:
                yield "{}{}{}{}".format(
                    name,
                    " {} ".format(self.sep) if self.sep != " " else self.sep,
                    value,
                    os.linesep,
                )
        for name, value in sections_dict.items():
            yield from value.gen_ini()

    def as_ini(self):
        return "".join(self.gen_ini())

    def as_dict(self):
        return dict(self)

    def dump(self):
        print(str(self))

    def __repr__(self, _repr_running=None):
        _repr_running = _repr_running or {}
        try:
            super_repr = super().__repr__(_repr_running)
        except TypeError:
            super_repr = super().__repr__()
        return os.linesep.join((super_repr, salt.utils.json.dumps(self, indent=4)))

    def __str__(self):
        return salt.utils.json.dumps(self, indent=4)

    def __eq__(self, item):
        return isinstance(item, self.__class__) and self.name == item.name

    def __ne__(self, item):
        return not (isinstance(item, self.__class__) and self.name == item.name)


class _Ini(_Section):
    def refresh(self, inicontents=None):
        if inicontents is None:
            if not os.path.exists(self.name):
                log.trace("File %s does not exist and will be created", self.name)
                return
            try:
                with salt.utils.files.fopen(self.name) as rfh:
                    inicontents = salt.utils.stringutils.to_unicode(rfh.read())
                    inicontents = os.linesep.join(inicontents.splitlines())
            except OSError as exc:
                if __opts__["test"] is False:
                    raise CommandExecutionError(
                        "Unable to open file '{}'. Exception: {}".format(self.name, exc)
                    )
        if not inicontents:
            return
        # Remove anything left behind from a previous run.
        self.clear()

        inicontents = INI_REGX.split(inicontents)
        inicontents.reverse()
        # Pop anything defined outside of a section (ie. at the top of
        # the ini file).
        super().refresh(inicontents.pop())
        for section_name, sect_ini in self._gen_tuples(inicontents):
            try:
                sect_obj = _Section(section_name, sect_ini, separator=self.sep)
                sect_obj.refresh()
                self.update({sect_obj.name: sect_obj})
            except StopIteration:
                pass

    def flush(self):
        try:
            with salt.utils.files.fopen(self.name, "wb") as outfile:
                ini_gen = self.gen_ini()
                next(ini_gen)
                ini_gen_list = list(ini_gen)
                # Avoid writing an initial line separator.
                if ini_gen_list:
                    ini_gen_list[0] = ini_gen_list[0].lstrip(os.linesep)
                outfile.writelines(salt.utils.data.encode(ini_gen_list))
        except OSError as exc:
            raise CommandExecutionError(
                "Unable to write file '{}'. Exception: {}".format(self.name, exc)
            )

    @staticmethod
    def get_ini_file(file_name, separator="="):
        inifile = _Ini(file_name, separator=separator)
        inifile.refresh()
        return inifile

    @staticmethod
    def _gen_tuples(list_object):
        while True:
            try:
                key = list_object.pop()
                value = list_object.pop()
            except IndexError:
                return
            else:
                yield key, value



"""
Print a stacktrace when sent a SIGUSR1 for debugging
"""

import inspect
import os
import signal
import sys
import tempfile
import time
import traceback

import salt.utils.files
import salt.utils.stringutils


def _makepretty(printout, stack):
    """
    Pretty print the stack trace and environment information
    for debugging those hard to reproduce user problems.  :)
    """
    printout.write("======== Salt Debug Stack Trace =========\n")
    traceback.print_stack(stack, file=printout)
    printout.write("=========================================\n")


def _handle_sigusr1(sig, stack):
    """
    Signal handler for SIGUSR1, only available on Unix-like systems
    """
    # When running in the foreground, do the right  thing
    # and spit out the debug info straight to the console
    if sys.stderr.isatty():
        output = sys.stderr
        _makepretty(output, stack)
    else:
        filename = "salt-debug-{}.log".format(int(time.time()))
        destfile = os.path.join(tempfile.gettempdir(), filename)
        with salt.utils.files.fopen(destfile, "w") as output:
            _makepretty(output, stack)


def _handle_sigusr2(sig, stack):
    """
    Signal handler for SIGUSR2, only available on Unix-like systems
    """
    try:
        import yappi
    except ImportError:
        return
    if yappi.is_running():
        yappi.stop()
        filename = "callgrind.salt-{}-{}".format(int(time.time()), os.getpid())
        destfile = os.path.join(tempfile.gettempdir(), filename)
        yappi.get_func_stats().save(destfile, type="CALLGRIND")
        if sys.stderr.isatty():
            sys.stderr.write("Saved profiling data to: {}\n".format(destfile))
        yappi.clear_stats()
    else:
        if sys.stderr.isatty():
            sys.stderr.write("Profiling started\n")
        yappi.start()


def enable_sig_handler(signal_name, handler):
    """
    Add signal handler for signal name if it exists on given platform
    """
    if hasattr(signal, signal_name):
        signal.signal(getattr(signal, signal_name), handler)


def enable_sigusr1_handler():
    """
    Pretty print a stack trace to the console or a debug log under /tmp
    when any of the salt daemons such as salt-master are sent a SIGUSR1
    """
    enable_sig_handler("SIGUSR1", _handle_sigusr1)
    # Also canonical BSD-way of printing progress is SIGINFO
    # which on BSD-derivatives can be sent via Ctrl+T
    enable_sig_handler("SIGINFO", _handle_sigusr1)


def enable_sigusr2_handler():
    """
    Toggle YAPPI profiler
    """
    enable_sig_handler("SIGUSR2", _handle_sigusr2)


def inspect_stack():
    """
    Return a string of which function we are currently in.
    """
    return {"co_name": inspect.stack()[1][3]}


def caller_name(skip=2, include_lineno=False):
    """
    Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed stack height

    Source: https://gist.github.com/techtonik/2151727
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ""
    parentframe = stack[start][0]

    name = []
    if include_lineno is True:
        try:
            lineno = inspect.getframeinfo(parentframe).lineno
        except:  # pylint: disable=bare-except
            lineno = None
    module = inspect.getmodule(parentframe)
    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)
    # detect classname
    if "self" in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parentframe.f_locals["self"].__class__.__name__)
    codename = parentframe.f_code.co_name
    if codename != "<module>":  # top level usually
        name.append(codename)  # function or a method
    del parentframe
    fullname = ".".join(name)
    if include_lineno and lineno:
        fullname += ":{}".format(lineno)
    return fullname



r"""
Module to manage Windows software repo on a Standalone Minion

``file_client: local`` must be set in the minion config file.

For documentation on Salt's Windows Repo feature, see :ref:`here
<windows-package-manager>`.
"""
import logging
import os

import salt.loader
import salt.output
import salt.template
import salt.utils.functools
import salt.utils.gitfs
import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltRenderError
from salt.runners.winrepo import genrepo as _genrepo
from salt.runners.winrepo import update_git_repos as _update_git_repos

# pylint: enable=unused-import
if salt.utils.platform.is_windows():
    _genrepo = salt.utils.functools.namespaced_function(_genrepo, globals())
    _update_git_repos = salt.utils.functools.namespaced_function(
        _update_git_repos, globals()
    )

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "winrepo"


def __virtual__():
    """
    Set the winrepo module if the OS is Windows
    """
    if salt.utils.platform.is_windows():
        return __virtualname__
    return (False, "This module only works on Windows.")


def _get_local_repo_dir(saltenv="base"):
    winrepo_source_dir = __opts__["winrepo_source_dir"]
    dirs = []
    dirs.append(salt.syspaths.CACHE_DIR)
    dirs.extend(["minion", "files"])
    dirs.append(saltenv)
    dirs.extend(winrepo_source_dir[7:].strip("/").split("/"))
    return os.sep.join(dirs)


def genrepo():
    r"""
    Generate winrepo_cachefile based on sls files in the winrepo_dir

    CLI Example:

    .. code-block:: bash

        salt-call winrepo.genrepo
    """
    return _genrepo(opts=__opts__, fire_event=False)


def update_git_repos(clean=False):
    """
    Checkout git repos containing :ref:`Windows Software Package Definitions
    <windows-package-manager>`.

    .. important::
        This function requires `Git for Windows`_ to be installed in order to
        work. When installing, make sure to select an installation option which
        permits the git executable to be run from the Command Prompt.

    .. _`Git for Windows`: https://git-for-windows.github.io/

    clean : False
        Clean repo cachedirs which are not configured under
        :conf_minion:`winrepo_remotes`.

        .. note::
            This option only applies if either pygit2_ or GitPython_ is
            installed into Salt's bundled Python.

        .. warning::
            This argument should not be set to ``True`` if a mix of git and
            non-git repo definitions are being used, as it will result in the
            non-git repo definitions being removed.

        .. versionadded:: 2015.8.0

        .. _GitPython: https://github.com/gitpython-developers/GitPython
        .. _pygit2: https://github.com/libgit2/pygit2

    CLI Example:

    .. code-block:: bash

        salt-call winrepo.update_git_repos
    """
    if not salt.utils.path.which("git"):
        raise CommandExecutionError(
            "Git for Windows is not installed, or not configured to be "
            "accessible from the Command Prompt"
        )
    return _update_git_repos(opts=__opts__, clean=clean, masterless=True)


def show_sls(name, saltenv="base"):
    r"""
    .. versionadded:: 2015.8.0

    Display the rendered software definition from a specific sls file in the
    local winrepo cache. This will parse all Jinja. Run pkg.refresh_db to pull
    the latest software definitions from the master.

    .. note::
        This function does not ask a master for an sls file to render. Instead
        it directly processes the file specified in `name`

    Args:
        name str: The name/path of the package you want to view. This can be the
        full path to a file on the minion file system or a file on the local
        minion cache.

        saltenv str: The default environment is ``base``

    Returns:
        dict: Returns a dictionary containing the rendered data structure

    .. note::
        To use a file from the minion cache start from the local winrepo root
        (``C:\salt\var\cache\salt\minion\files\base\win\repo-ng``). If you have
        ``.sls`` files organized in subdirectories you'll have to denote them
        with ``.``. For example, if you have a ``test`` directory in the winrepo
        root with a ``gvim.sls`` file inside, would target that file like so:
        ``test.gvim``. Directories can be targeted as well as long as they
        contain an ``init.sls`` inside. For example, if you have a ``node``
        directory with an ``init.sls`` inside, target that like so: ``node``.

    CLI Example:

    .. code-block:: bash

        salt '*' winrepo.show_sls gvim
        salt '*' winrepo.show_sls test.npp
        salt '*' winrepo.show_sls C:\test\gvim.sls
    """
    # Passed a filename
    if os.path.exists(name):
        sls_file = name

    # Use a winrepo path
    else:
        # Get the location of the local repo
        repo = _get_local_repo_dir(saltenv)

        # Add the sls file name to the path
        repo = repo.split("\\")
        definition = name.split(".")
        repo.extend(definition)

        # Check for the sls file by name
        sls_file = "{}.sls".format(os.sep.join(repo))
        if not os.path.exists(sls_file):

            # Maybe it's a directory with an init.sls
            sls_file = "{}\\init.sls".format(os.sep.join(repo))
            if not os.path.exists(sls_file):

                # It's neither, return
                return "Software definition {} not found".format(name)

    # Load the renderer
    renderers = salt.loader.render(__opts__, __salt__)
    config = {}

    # Run the file through the renderer
    try:
        config = salt.template.compile_template(
            sls_file,
            renderers,
            __opts__["renderer"],
            __opts__["renderer_blacklist"],
            __opts__["renderer_whitelist"],
        )

    # Return the error if any
    except SaltRenderError as exc:
        log.debug("Failed to compile %s.", sls_file)
        log.debug("Error: %s.", exc)
        config["Message"] = "Failed to compile {}".format(sls_file)
        config["Error"] = "{}".format(exc)

    return config



"""
Execution module to manage Cisco Nexus Switches (NX-OS) over the NX-API

.. versionadded:: 2019.2.0

Execution module used to interface the interaction with a remote or local Nexus
switch whether we're running in a Proxy Minion or regular Minion (or regular
Minion running directly on the Nexus switch).

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:platform:   any

.. note::

    To be able to use this module you need to enable to NX-API on your switch,
    by executing ``feature nxapi`` in configuration mode.

    Configuration example:

    .. code-block:: bash

        switch# conf t
        switch(config)# feature nxapi

    To check that NX-API is properly enabled, execute ``show nxapi``.

    Output example:

    .. code-block:: bash

        switch# show nxapi
        nxapi enabled
        HTTPS Listen on port 443

.. note::

    NX-API requires modern NXOS distributions, typically at least 7.0 depending
    on the hardware. Due to reliability reasons it is recommended to run the
    most recent version.

    Check https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus7000/sw/programmability/guide/b_Cisco_Nexus_7000_Series_NX-OS_Programmability_Guide/b_Cisco_Nexus_7000_Series_NX-OS_Programmability_Guide_chapter_0101.html
    for more details.

Usage
-----

This module can equally be used via the :mod:`nxos_api<salt.proxy.nxos_api>`
Proxy module or directly from an arbitrary (Proxy) Minion that is running on a
machine having access to the network device API. Given that there are no
external dependencies, this module can very well used when using the regular
Salt Minion directly installed on the switch.

When running outside of the :mod:`nxos_api Proxy<salt.proxy.nxos_api>`
(i.e., from another Proxy Minion type, or regular Minion), the NX-API connection
arguments can be either specified from the CLI when executing the command, or
in a configuration block under the ``nxos_api`` key in the configuration opts
(i.e., (Proxy) Minion configuration file), or Pillar. The module supports these
simultaneously. These fields are the exact same supported by the ``nxos_api``
Proxy Module:

transport: ``https``
    Specifies the type of connection transport to use. Valid values for the
    connection are ``http``, and  ``https``.

host: ``localhost``
    The IP address or DNS host name of the connection device.

username: ``admin``
    The username to pass to the device to authenticate the NX-API connection.

password
    The password to pass to the device to authenticate the NX-API connection.

port
    The TCP port of the endpoint for the NX-API connection. If this keyword is
    not specified, the default value is automatically determined by the
    transport type (``80`` for ``http``, or ``443`` for ``https``).

timeout: ``60``
    Time in seconds to wait for the device to respond. Default: 60 seconds.

verify: ``True``
    Either a boolean, in which case it controls whether we verify the NX-API
    TLS certificate, or a string, in which case it must be a path to a CA bundle
    to use. Defaults to ``True``.

    When there is no certificate configuration on the device and this option is
    set as ``True`` (default), the commands will fail with the following error:
    ``SSLError: [SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed (_ssl.c:581)``.
    In this case, you either need to configure a proper certificate on the
    device (*recommended*), or bypass the checks setting this argument as ``False``
    with all the security risks considered.

    Check https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus3000/sw/programmability/6_x/b_Cisco_Nexus_3000_Series_NX-OS_Programmability_Guide/b_Cisco_Nexus_3000_Series_NX-OS_Programmability_Guide_chapter_01.html
    to see how to properly configure the certificate.

Example (when not running in a ``nxos_api`` Proxy Minion):

.. code-block:: yaml

  nxos_api:
    username: test
    password: test

In case the ``username`` and ``password`` are the same on any device you are
targeting, the block above (besides other parameters specific to your
environment you might need) should suffice to be able to execute commands from
outside a ``nxos_api`` Proxy, e.g.:

.. code-block:: bash

    salt-call --local nxos_api.show 'show lldp neighbors' raw_text
    # The command above is available when running in a regular Minion where Salt is installed

    salt '*' nxos_api.show 'show version' raw_text=False

.. note::

    Remember that the above applies only when not running in a ``nxos_api`` Proxy
    Minion. If you want to use the :mod:`nxos_api Proxy<salt.proxy.nxos_api>`,
    please follow the documentation notes for a proper setup.
"""

import difflib
import logging

from salt.exceptions import CommandExecutionError, SaltException

# -----------------------------------------------------------------------------
# execution module properties
# -----------------------------------------------------------------------------

__proxyenabled__ = ["*"]
# Any Proxy Minion should be able to execute these

__virtualname__ = "nxos_api"
# The Execution Module will be identified as ``nxos_api``
# The ``nxos`` namespace is already taken, used for SSH-based connections.

# -----------------------------------------------------------------------------
# globals
# -----------------------------------------------------------------------------

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# propery functions
# -----------------------------------------------------------------------------


def __virtual__():
    """
    This module does not have external dependencies, hence it is widely
    available.
    """
    # No extra requirements, uses Salt native modules.
    return __virtualname__


# -----------------------------------------------------------------------------
# helper functions
# -----------------------------------------------------------------------------


def _cli_command(commands, method="cli", **kwargs):
    """
    Execute a list of CLI commands.
    """
    if not isinstance(commands, (list, tuple)):
        commands = [commands]
    rpc_responses = rpc(commands, method=method, **kwargs)
    txt_responses = []
    for rpc_reponse in rpc_responses:
        error = rpc_reponse.get("error")
        if error:
            cmd = rpc_reponse.get("command")
            if "data" in error:
                msg = 'The command "{cmd}" raised the error "{err}".'.format(
                    cmd=cmd, err=error["data"]["msg"]
                )
                raise SaltException(msg)
            else:
                msg = 'Invalid command: "{cmd}".'.format(cmd=cmd)
                raise SaltException(msg)
        txt_responses.append(rpc_reponse["result"])
    return txt_responses


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def rpc(commands, method="cli", **kwargs):
    """
    Execute an arbitrary RPC request via the Nexus API.

    commands
        The commands to be executed.

    method: ``cli``
        The type of the response, i.e., raw text (``cli_ascii``) or structured
        document (``cli``). Defaults to ``cli`` (structured data).

    transport: ``https``
        Specifies the type of connection transport to use. Valid values for the
        connection are ``http``, and  ``https``.

    host: ``localhost``
        The IP address or DNS host name of the connection device.

    username: ``admin``
        The username to pass to the device to authenticate the NX-API connection.

    password
        The password to pass to the device to authenticate the NX-API connection.

    port
        The TCP port of the endpoint for the NX-API connection. If this keyword is
        not specified, the default value is automatically determined by the
        transport type (``80`` for ``http``, or ``443`` for ``https``).

    timeout: ``60``
        Time in seconds to wait for the device to respond. Default: 60 seconds.

    verify: ``True``
        Either a boolean, in which case it controls whether we verify the NX-API
        TLS certificate, or a string, in which case it must be a path to a CA bundle
        to use. Defaults to ``True``.

    CLI Example:

    .. code-block:: bash

        salt-call --local nxos_api.rpc 'show version'
    """
    nxos_api_kwargs = __salt__["config.get"]("nxos_api", {})
    nxos_api_kwargs.update(**kwargs)
    if (
        "nxos_api.rpc" in __proxy__
        and __salt__["config.get"]("proxy:proxytype") == "nxos_api"
    ):
        # If the nxos_api.rpc Proxy function is available and currently running
        # in a nxos_api Proxy Minion
        return __proxy__["nxos_api.rpc"](commands, method=method, **nxos_api_kwargs)
    nxos_api_kwargs = __salt__["config.get"]("nxos_api", {})
    nxos_api_kwargs.update(**kwargs)
    return __utils__["nxos_api.rpc"](commands, method=method, **nxos_api_kwargs)


def show(commands, raw_text=True, **kwargs):
    """
    Execute one or more show (non-configuration) commands.

    commands
        The commands to be executed.  Multiple commands should
        be specified as a list.

    raw_text: ``True``
        Whether to return raw text or structured data.

    transport: ``https``
        Specifies the type of connection transport to use. Valid values for the
        connection are ``http``, and  ``https``.

    host: ``localhost``
        The IP address or DNS host name of the connection device.

    username: ``admin``
        The username to pass to the device to authenticate the NX-API connection.

    password
        The password to pass to the device to authenticate the NX-API connection.

    port
        The TCP port of the endpoint for the NX-API connection. If this keyword is
        not specified, the default value is automatically determined by the
        transport type (``80`` for ``http``, or ``443`` for ``https``).

    timeout: ``60``
        Time in seconds to wait for the device to respond. Default: 60 seconds.

    verify: ``True``
        Either a boolean, in which case it controls whether we verify the NX-API
        TLS certificate, or a string, in which case it must be a path to a CA bundle
        to use. Defaults to ``True``.

    CLI Example:

    .. code-block:: bash

        salt-call --local nxos_api.show 'show version'
        salt '*' nxos_api.show "['show bgp sessions','show processes']" raw_text=False
        salt 'regular-minion' nxos_api.show 'show interfaces' host=sw01.example.com username=test password=test
    """
    ret = []
    if raw_text:
        method = "cli_ascii"
        key = "msg"
    else:
        method = "cli"
        key = "body"
    response_list = _cli_command(commands, method=method, **kwargs)
    ret = [response[key] for response in response_list if response]
    return ret


def config(
    commands=None,
    config_file=None,
    template_engine="jinja",
    context=None,
    defaults=None,
    saltenv="base",
    **kwargs
):
    """
    Configures the Nexus switch with the specified commands.

    This method is used to send configuration commands to the switch.  It
    will take either a string or a list and prepend the necessary commands
    to put the session into config mode.

    .. warning::

        All the commands will be applied directly into the running-config.

    config_file
        The source file with the configuration commands to be sent to the
        device.

        The file can also be a template that can be rendered using the template
        engine of choice.

        This can be specified using the absolute path to the file, or using one
        of the following URL schemes:

        - ``salt://``, to fetch the file from the Salt fileserver.
        - ``http://`` or ``https://``
        - ``ftp://``
        - ``s3://``
        - ``swift://``

    commands
        The commands to send to the switch in config mode.  If the commands
        argument is a string it will be cast to a list.
        The list of commands will also be prepended with the necessary commands
        to put the session in config mode.

        .. note::

            This argument is ignored when ``config_file`` is specified.

    template_engine: ``jinja``
        The template engine to use when rendering the source file. Default:
        ``jinja``. To simply fetch the file without attempting to render, set
        this argument to ``None``.

    context
        Variables to add to the template context.

    defaults
        Default values of the context_dict.

    transport: ``https``
        Specifies the type of connection transport to use. Valid values for the
        connection are ``http``, and  ``https``.

    host: ``localhost``
        The IP address or DNS host name of the connection device.

    username: ``admin``
        The username to pass to the device to authenticate the NX-API connection.

    password
        The password to pass to the device to authenticate the NX-API connection.

    port
        The TCP port of the endpoint for the NX-API connection. If this keyword is
        not specified, the default value is automatically determined by the
        transport type (``80`` for ``http``, or ``443`` for ``https``).

    timeout: ``60``
        Time in seconds to wait for the device to respond. Default: 60 seconds.

    verify: ``True``
        Either a boolean, in which case it controls whether we verify the NX-API
        TLS certificate, or a string, in which case it must be a path to a CA bundle
        to use. Defaults to ``True``.

    CLI Example:

    .. code-block:: bash

        salt '*' nxos_api.config commands="['spanning-tree mode mstp']"
        salt '*' nxos_api.config config_file=salt://config.txt
        salt '*' nxos_api.config config_file=https://bit.ly/2LGLcDy context="{'servers': ['1.2.3.4']}"
    """
    initial_config = show("show running-config", **kwargs)[0]
    if config_file:
        file_str = __salt__["cp.get_file_str"](config_file, saltenv=saltenv)
        if file_str is False:
            raise CommandExecutionError("Source file {} not found".format(config_file))
    elif commands:
        if isinstance(commands, str):
            commands = [commands]
        file_str = "\n".join(commands)
        # unify all the commands in a single file, to render them in a go
    if template_engine:
        file_str = __salt__["file.apply_template_on_contents"](
            file_str, template_engine, context, defaults, saltenv
        )
    # whatever the source of the commands would be, split them line by line
    commands = [line for line in file_str.splitlines() if line.strip()]
    ret = _cli_command(commands, **kwargs)
    current_config = show("show running-config", **kwargs)[0]
    diff = difflib.unified_diff(
        initial_config.splitlines(1)[4:], current_config.splitlines(1)[4:]
    )
    return "".join([x.replace("\r", "") for x in diff])



"""
Module for running arbitrary tests
"""

__virtualname__ = "test"


def __virtual__():
    return __virtualname__


def recho(text):
    """
    Return a reversed string

    CLI Example:

    .. code-block:: bash

        salt '*' test.recho 'foo bar baz quo qux'
    """
    return text[::-1]



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


import pytest

import salt.modules.napalm_bgp as napalm_bgp
import tests.support.napalm as napalm_test_support
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        "__salt__": {
            "config.get": MagicMock(
                return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
            ),
            "file.file_exists": napalm_test_support.true,
            "file.join": napalm_test_support.join,
            "file.get_managed": napalm_test_support.get_managed_file,
            "random.hash": napalm_test_support.random_hash,
        },
    }

    return {napalm_bgp: module_globals}


def test_config():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_bgp.config("test_group")
        assert ret["out"] == napalm_test_support.TEST_BGP_CONFIG.copy()


def test_neighbors():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_bgp.neighbors("test_address")
        assert ret["out"] == napalm_test_support.TEST_BGP_NEIGHBORS.copy()



"""
Module for Solaris 10's zoneadm

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:platform:      OmniOS,OpenIndiana,SmartOS,OpenSolaris,Solaris 10

.. versionadded:: 2017.7.0

.. warning::
    Oracle Solaris 11's zoneadm is not supported by this module!
"""

import logging

import salt.utils.decorators
import salt.utils.path

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "zoneadm"

# Function aliases
__func_alias__ = {"list_zones": "list"}


@salt.utils.decorators.memoize
def _is_globalzone():
    """
    Check if we are running in the globalzone
    """
    if not __grains__["kernel"] == "SunOS":
        return False

    zonename = __salt__["cmd.run_all"]("zonename")
    if zonename["retcode"]:
        return False
    if zonename["stdout"] == "global":
        return True

    return False


def _is_uuid(zone):
    """
    Check if zone is actually a UUID
    """
    return len(zone) == 36 and zone.index("-") == 8


def __virtual__():
    """
    We are available if we are have zoneadm and are the global zone on
    Solaris 10, OmniOS, OpenIndiana, OpenSolaris, or Smartos.
    """
    if _is_globalzone() and salt.utils.path.which("zoneadm"):
        if __grains__["os"] in ["OpenSolaris", "SmartOS", "OmniOS", "OpenIndiana"]:
            return __virtualname__
        elif (
            __grains__["os"] == "Oracle Solaris"
            and int(__grains__["osmajorrelease"]) == 10
        ):
            return __virtualname__
    return (
        False,
        "{} module can only be loaded in a solaris globalzone.".format(__virtualname__),
    )


def list_zones(verbose=True, installed=False, configured=False, hide_global=True):
    """
    List all zones

    verbose : boolean
        display additional zone information
    installed : boolean
        include installed zones in output
    configured : boolean
        include configured zones in output
    hide_global : boolean
        do not include global zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.list
    """
    zones = {}

    ## fetch zones
    header = "zoneid:zonename:state:zonepath:uuid:brand:ip-type".split(":")
    zone_data = __salt__["cmd.run_all"]("zoneadm list -p -c")
    if zone_data["retcode"] == 0:
        for zone in zone_data["stdout"].splitlines():
            zone = zone.split(":")

            # create zone_t
            zone_t = {}
            for num, val in enumerate(header):
                zone_t[val] = zone[num]

            # skip if global and hide_global
            if hide_global and zone_t["zonename"] == "global":
                continue

            # skip installed and configured
            if not installed and zone_t["state"] == "installed":
                continue
            if not configured and zone_t["state"] == "configured":
                continue

            # update dict
            zones[zone_t["zonename"]] = zone_t
            del zones[zone_t["zonename"]]["zonename"]

    return zones if verbose else sorted(zones.keys())


def boot(zone, single=False, altinit=None, smf_options=None):
    """
    Boot (or activate) the specified zone.

    zone : string
        name or uuid of the zone
    single : boolean
        boots only to milestone svc:/milestone/single-user:default.
    altinit : string
        valid path to an alternative executable to be the primordial process.
    smf_options : string
        include two categories of options to control booting behavior of
        the service management facility: recovery options and messages options.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.boot clementine
        salt '*' zoneadm.boot maeve single=True
        salt '*' zoneadm.boot teddy single=True smf_options=verbose
    """
    ret = {"status": True}

    ## build boot_options
    boot_options = ""
    if single:
        boot_options = "-s {}".format(boot_options)
    if altinit:  # note: we cannot validate the path, as this is local to the zonepath.
        boot_options = "-i {} {}".format(altinit, boot_options)
    if smf_options:
        boot_options = "-m {} {}".format(smf_options, boot_options)
    if boot_options != "":
        boot_options = " -- {}".format(boot_options.strip())

    ## execute boot
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} boot{boot_opts}".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
            boot_opts=boot_options,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def reboot(zone, single=False, altinit=None, smf_options=None):
    """
    Restart the zone. This is equivalent to a halt boot sequence.

    zone : string
        name or uuid of the zone
    single : boolean
        boots only to milestone svc:/milestone/single-user:default.
    altinit : string
        valid path to an alternative executable to be the primordial process.
    smf_options : string
        include two categories of options to control booting behavior of
        the service management facility: recovery options and messages options.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.reboot dolores
        salt '*' zoneadm.reboot teddy single=True
    """
    ret = {"status": True}

    ## build boot_options
    boot_options = ""
    if single:
        boot_options = "-s {}".format(boot_options)
    if altinit:  # note: we cannot validate the path, as this is local to the zonepath.
        boot_options = "-i {} {}".format(altinit, boot_options)
    if smf_options:
        boot_options = "-m {} {}".format(smf_options, boot_options)
    if boot_options != "":
        boot_options = " -- {}".format(boot_options.strip())

    ## execute reboot
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} reboot{boot_opts}".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
            boot_opts=boot_options,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def halt(zone):
    """
    Halt the specified zone.

    zone : string
        name or uuid of the zone

    .. note::
        To cleanly shutdown the zone use the shutdown function.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.halt hector
    """
    ret = {"status": True}

    ## halt zone
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} halt".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def shutdown(zone, reboot=False, single=False, altinit=None, smf_options=None):
    """
    Gracefully shutdown the specified zone.

    zone : string
        name or uuid of the zone
    reboot : boolean
        reboot zone after shutdown (equivalent of shutdown -i6 -g0 -y)
    single : boolean
        boots only to milestone svc:/milestone/single-user:default.
    altinit : string
        valid path to an alternative executable to be the primordial process.
    smf_options : string
        include two categories of options to control booting behavior of
        the service management facility: recovery options and messages options.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.shutdown peter
        salt '*' zoneadm.shutdown armistice reboot=True
    """
    ret = {"status": True}

    ## build boot_options
    boot_options = ""
    if single:
        boot_options = "-s {}".format(boot_options)
    if altinit:  # note: we cannot validate the path, as this is local to the zonepath.
        boot_options = "-i {} {}".format(altinit, boot_options)
    if smf_options:
        boot_options = "-m {} {}".format(smf_options, boot_options)
    if boot_options != "":
        boot_options = " -- {}".format(boot_options.strip())

    ## shutdown zone
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} shutdown{reboot}{boot_opts}".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
            reboot=" -r" if reboot else "",
            boot_opts=boot_options,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def detach(zone):
    """
    Detach the specified zone.

    zone : string
        name or uuid of the zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.detach kissy
    """
    ret = {"status": True}

    ## detach zone
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} detach".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def attach(zone, force=False, brand_opts=None):
    """
    Attach the specified zone.

    zone : string
        name of the zone
    force : boolean
        force the zone into the "installed" state with no validation
    brand_opts : string
        brand specific options to pass

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.attach lawrence
        salt '*' zoneadm.attach lawrence True
    """
    ret = {"status": True}

    ## attach zone
    res = __salt__["cmd.run_all"](
        "zoneadm -z {zone} attach{force}{brand_opts}".format(
            zone=zone,
            force=" -F" if force else "",
            brand_opts=" {}".format(brand_opts) if brand_opts else "",
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def ready(zone):
    """
    Prepares a zone for running applications.

    zone : string
        name or uuid of the zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.ready clementine
    """
    ret = {"status": True}

    ## ready zone
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} ready".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def verify(zone):
    """
    Check to make sure the configuration of the specified
    zone can safely be installed on the machine.

    zone : string
        name of the zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.verify dolores
    """
    ret = {"status": True}

    ## verify zone
    res = __salt__["cmd.run_all"](
        "zoneadm -z {zone} verify".format(
            zone=zone,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def move(zone, zonepath):
    """
    Move zone to new zonepath.

    zone : string
        name or uuid of the zone
    zonepath : string
        new zonepath

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.move meave /sweetwater/meave
    """
    ret = {"status": True}

    ## verify zone
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} move {path}".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
            path=zonepath,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def uninstall(zone):
    """
    Uninstall the specified zone from the system.

    zone : string
        name or uuid of the zone

    .. warning::
        The -F flag is always used to avoid the prompts when uninstalling.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.uninstall teddy
    """
    ret = {"status": True}

    ## uninstall zone
    res = __salt__["cmd.run_all"](
        "zoneadm {zone} uninstall -F".format(
            zone="-u {}".format(zone) if _is_uuid(zone) else "-z {}".format(zone),
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def install(zone, nodataset=False, brand_opts=None):
    """
    Install the specified zone from the system.

    zone : string
        name of the zone
    nodataset : boolean
        do not create a ZFS file system
    brand_opts : string
        brand specific options to pass

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.install dolores
        salt '*' zoneadm.install teddy True
    """
    ret = {"status": True}

    ## install zone
    res = __salt__["cmd.run_all"](
        "zoneadm -z {zone} install{nodataset}{brand_opts}".format(
            zone=zone,
            nodataset=" -x nodataset" if nodataset else "",
            brand_opts=" {}".format(brand_opts) if brand_opts else "",
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret


def clone(zone, source, snapshot=None):
    """
    Install a zone by copying an existing installed zone.

    zone : string
        name of the zone
    source : string
        zone to clone from
    snapshot : string
        optional name of snapshot to use as source

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.clone clementine dolores
    """
    ret = {"status": True}

    ## install zone
    res = __salt__["cmd.run_all"](
        "zoneadm -z {zone} clone {snapshot}{source}".format(
            zone=zone,
            source=source,
            snapshot="-s {} ".format(snapshot) if snapshot else "",
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    ret["message"] = ret["message"].replace("zoneadm: ", "")
    if ret["message"] == "":
        del ret["message"]

    return ret



"""
Support for GRUB Legacy
"""

import os

import salt.utils.decorators as decorators
import salt.utils.files
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "grub"


def __virtual__():
    """
    Only load the module if grub is installed
    """
    if os.path.exists(_detect_conf()):
        return __virtualname__
    return (
        False,
        "The grub_legacy execution module cannot be loaded: "
        "the grub config file does not exist in /boot/grub/",
    )


@decorators.memoize
def _detect_conf():
    """
    GRUB conf location differs depending on distro
    """
    if __grains__["os_family"] == "RedHat":
        return "/boot/grub/grub.conf"
    # Defaults for Ubuntu, Debian, Arch, and others
    return "/boot/grub/menu.lst"


def version():
    """
    Return server version from grub --version

    CLI Example:

    .. code-block:: bash

        salt '*' grub.version
    """
    cmd = "/sbin/grub --version"
    out = __salt__["cmd.run"](cmd)
    return out


def conf():
    """
    Parse GRUB conf file

    CLI Example:

    .. code-block:: bash

        salt '*' grub.conf
    """
    stanza = ""
    stanzas = []
    in_stanza = False
    ret = {}
    pos = 0
    try:
        with salt.utils.files.fopen(_detect_conf(), "r") as _fp:
            for line in _fp:
                line = salt.utils.stringutils.to_unicode(line)
                if line.startswith("#"):
                    continue
                if line.startswith("\n"):
                    in_stanza = False
                    if "title" in stanza:
                        stanza += "order {}".format(pos)
                        pos += 1
                        stanzas.append(stanza)
                    stanza = ""
                    continue
                if line.strip().startswith("title"):
                    if in_stanza:
                        stanza += "order {}".format(pos)
                        pos += 1
                        stanzas.append(stanza)
                        stanza = ""
                    else:
                        in_stanza = True
                if in_stanza:
                    stanza += line
                if not in_stanza:
                    key, value = _parse_line(line)
                    ret[key] = value
            if in_stanza:
                if not line.endswith("\n"):
                    line += "\n"
                stanza += line
                stanza += "order {}".format(pos)
                pos += 1
                stanzas.append(stanza)
    except OSError as exc:
        msg = "Could not read grub config: {0}"
        raise CommandExecutionError(msg.format(exc))

    ret["stanzas"] = []
    for stanza in stanzas:
        mydict = {}
        for line in stanza.strip().splitlines():
            key, value = _parse_line(line)
            mydict[key] = value
        ret["stanzas"].append(mydict)
    return ret


def _parse_line(line=""):
    """
    Used by conf() to break config lines into
    name/value pairs
    """
    parts = line.split()
    key = parts.pop(0)
    value = " ".join(parts)
    return key, value



"""
Oracle DataBase connection module

:maintainer: Vladimir Bormotov <bormotov@gmail.com>

:maturity: new

:depends: cx_Oracle

:platform: all

:configuration: module provide connections for multiple Oracle DB instances.

    **OS Environment**

    .. code-block:: yaml

        ORACLE_HOME: path to oracle product
        PATH: path to Oracle Client libs need to be in PATH

    **pillar**

    .. code-block:: yaml

        oracle:
          dbs:
            <db>:
              uri: connection credentials in format:
            user/password@host[:port]/sid[ servicename as {sysdba|sysoper}]
              optional keyword servicename will determine whether it is a sid or service_name
            <db>:
              uri: .....
"""

import logging
import os

from salt.utils.decorators import depends

log = logging.getLogger(__name__)

try:
    import cx_Oracle

    MODE = {"sysdba": cx_Oracle.SYSDBA, "sysoper": cx_Oracle.SYSOPER}
    HAS_CX_ORACLE = True
except ImportError:
    MODE = {"sysdba": 2, "sysoper": 4}
    HAS_CX_ORACLE = False

__virtualname__ = "oracle"


def __virtual__():
    """
    Load module only if cx_Oracle installed
    """
    if HAS_CX_ORACLE:
        return __virtualname__
    return (
        False,
        "oracle execution module not loaded: python oracle library not found.",
    )


def _cx_oracle_req():
    """
    Fallback function stub
    """
    return 'Need "cx_Oracle" and Oracle Client installed for this function exist'


def _unicode_output(cursor, name, default_type, size, precision, scale):
    """
    Return strings values as python unicode string

    http://www.oracle.com/technetwork/articles/dsl/tuininga-cx-oracle-084866.html
    """
    if default_type in (
        cx_Oracle.STRING,
        cx_Oracle.LONG_STRING,
        cx_Oracle.FIXED_CHAR,
        cx_Oracle.CLOB,
    ):
        return cursor.var(str, size, cursor.arraysize)


def _connect(uri):
    """
    uri = user/password@host[:port]/sid[ as {sysdba|sysoper}]

    Return cx_Oracle.Connection instance
    """
    # cx_Oracle.Connection() not support 'as sysdba' syntax
    uri_l = uri.rsplit(" as ", 1)
    if len(uri_l) == 2:
        credentials, mode = uri_l
        mode = MODE[mode]
    else:
        credentials = uri_l[0]
        mode = 0

    serv_name = False
    userpass, hostportsid = credentials.split("@")
    user, password = userpass.split("/")
    hostport, sid = hostportsid.split("/")
    if "servicename" in sid:
        serv_name = True
        sid = sid.split("servicename")[0].strip()
    hostport_l = hostport.split(":")
    if len(hostport_l) == 2:
        host, port = hostport_l
    else:
        host = hostport_l[0]
        port = 1521
    log.debug("connect: %s", (user, password, host, port, sid, mode))
    # force UTF-8 client encoding
    os.environ["NLS_LANG"] = ".AL32UTF8"
    if serv_name:
        conn = cx_Oracle.connect(
            user, password, cx_Oracle.makedsn(host, port, service_name=sid), mode
        )
    else:
        conn = cx_Oracle.connect(
            user, password, cx_Oracle.makedsn(host, port, sid), mode
        )
    conn.outputtypehandler = _unicode_output
    return conn


@depends("cx_Oracle", fallback_function=_cx_oracle_req)
def run_query(db, query):
    """
    Run SQL query and return result

    CLI Example:

    .. code-block:: bash

        salt '*' oracle.run_query my_db "select * from my_table"
    """
    log.debug("run query on %s: %s", db, query)
    conn = _connect(show_dbs(db)[db]["uri"])
    return conn.cursor().execute(query).fetchall()


def show_dbs(*dbs):
    """
    Show databases configuration from pillar. Filter by `*args`

    CLI Example:

    .. code-block:: bash

        salt '*' oracle.show_dbs
        salt '*' oracle.show_dbs my_db
    """
    if dbs:
        log.debug("get dbs from pillar: %s", dbs)
        result = {}
        for db in dbs:
            result[db] = __salt__["pillar.get"]("oracle:dbs:" + db)
        return result
    else:
        pillar_dbs = __salt__["pillar.get"]("oracle:dbs")
        log.debug("get all (%s) dbs from pillar", len(pillar_dbs))
        return pillar_dbs


@depends("cx_Oracle", fallback_function=_cx_oracle_req)
def version(*dbs):
    """
    Server Version (select banner  from v$version)

    CLI Example:

    .. code-block:: bash

        salt '*' oracle.version
        salt '*' oracle.version my_db
    """
    pillar_dbs = __salt__["pillar.get"]("oracle:dbs")
    get_version = lambda x: [
        r[0] for r in run_query(x, "select banner from v$version order by banner")
    ]
    result = {}
    if dbs:
        log.debug("get db versions for: %s", dbs)
        for db in dbs:
            if db in pillar_dbs:
                result[db] = get_version(db)
    else:
        log.debug("get all (%s) dbs versions", len(dbs))
        for db in dbs:
            result[db] = get_version(db)
    return result


@depends("cx_Oracle", fallback_function=_cx_oracle_req)
def client_version():
    """
    Oracle Client Version

    CLI Example:

    .. code-block:: bash

        salt '*' oracle.client_version
    """
    return ".".join(str(x) for x in cx_Oracle.clientversion())


def show_pillar(item=None):
    """
    Show Pillar segment oracle.* and subitem with notation "item:subitem"

    CLI Example:

    .. code-block:: bash

        salt '*' oracle.show_pillar
        salt '*' oracle.show_pillar dbs:my_db
    """
    if item:
        return __salt__["pillar.get"]("oracle:" + item)
    else:
        return __salt__["pillar.get"]("oracle")


def show_env():
    """
    Show Environment used by Oracle Client

    CLI Example:

    .. code-block:: bash

        salt '*' oracle.show_env

    .. note::
        at first _connect() ``NLS_LANG`` will forced to '.AL32UTF8'
    """
    envs = ["PATH", "ORACLE_HOME", "TNS_ADMIN", "NLS_LANG"]
    result = {}
    for env in envs:
        if env in os.environ:
            result[env] = os.environ[env]
    return result



"""
Module for viewing and modifying OpenBSD sysctl parameters
"""

import os
import re

import salt.utils.data
import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "sysctl"


def __virtual__():
    """
    Only run on OpenBSD systems
    """
    if __grains__["os"] == "OpenBSD":
        return __virtualname__
    return (
        False,
        "The openbsd_sysctl execution module cannot be loaded: "
        "only available on OpenBSD systems.",
    )


def show(config_file=False):
    """
    Return a list of sysctl parameters for this minion

    config: Pull the data from the system configuration file
        instead of the live data.

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.show
    """
    cmd = "sysctl"
    ret = {}
    out = __salt__["cmd.run_stdout"](cmd, output_loglevel="trace")
    for line in out.splitlines():
        if not line or "=" not in line:
            continue
        comps = line.split("=", 1)
        ret[comps[0]] = comps[1]
    return ret


def get(name):
    """
    Return a single sysctl parameter for this minion

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.get hw.physmem
    """
    cmd = "sysctl -n {}".format(name)
    out = __salt__["cmd.run"](cmd)
    return out


def assign(name, value):
    """
    Assign a single sysctl parameter for this minion

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.assign net.inet.ip.forwarding 1
    """
    ret = {}
    cmd = 'sysctl {}="{}"'.format(name, value)
    data = __salt__["cmd.run_all"](cmd)

    # Certain values cannot be set from this console, at the current
    # securelevel or there are other restrictions that prevent us
    # from applying the setting rightaway.
    if (
        re.match(r"^sysctl:.*: Operation not permitted$", data["stderr"])
        or data["retcode"] != 0
    ):
        raise CommandExecutionError("sysctl failed: {}".format(data["stderr"]))
    new_name, new_value = data["stdout"].split(":", 1)
    ret[new_name] = new_value.split(" -> ")[-1]
    return ret


def persist(name, value, config="/etc/sysctl.conf"):
    """
    Assign and persist a simple sysctl parameter for this minion

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.persist net.inet.ip.forwarding 1
    """
    nlines = []
    edited = False
    value = str(value)

    # create /etc/sysctl.conf if not present
    if not os.path.isfile(config):
        try:
            with salt.utils.files.fopen(config, "w+"):
                pass
        except OSError:
            msg = "Could not create {0}"
            raise CommandExecutionError(msg.format(config))

    with salt.utils.files.fopen(config, "r") as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)
            if not line.startswith("{}=".format(name)):
                nlines.append(line)
                continue
            else:
                key, rest = line.split("=", 1)
                if rest.startswith('"'):
                    _, rest_v, rest = rest.split('"', 2)
                elif rest.startswith("'"):
                    _, rest_v, rest = rest.split("'", 2)
                else:
                    rest_v = rest.split()[0]
                    rest = rest[len(rest_v) :]
                if rest_v == value:
                    return "Already set"
                new_line = "{}={}{}".format(key, value, rest)
                nlines.append(new_line)
                edited = True
    if not edited:
        nlines.append("{}={}\n".format(name, value))
    with salt.utils.files.fopen(config, "wb") as ofile:
        ofile.writelines(salt.utils.data.encode(nlines))

    assign(name, value)
    return "Updated"



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import logging

import pytest  # pylint: disable=unused-import

from salt.utils.decorators import state as statedecorators

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.core_test,
]


def test_sub_state_output_check_changes_is_dict():
    """
    Test that changes key contains a dictionary.
    :return:
    """
    data = {"changes": {}, "sub_state_run": [{"changes": []}]}
    out = statedecorators.OutputUnifier("content_check")(lambda: data)()
    assert "'Changes' should be a dictionary" in out["sub_state_run"][0]["comment"]
    assert not out["sub_state_run"][0]["result"]


def test_sub_state_output_check_return_is_dict():
    """
    Test for the entire return is a dictionary
    :return:
    """
    data = {"sub_state_run": [["whatever"]]}
    out = statedecorators.OutputUnifier("content_check")(lambda: data)()
    assert (
        "Malformed state return. Data must be a dictionary type"
        in out["sub_state_run"][0]["comment"]
    )
    assert not out["sub_state_run"][0]["result"]


def test_sub_state_output_check_return_has_nrc():
    """
    Test for name/result/comment keys are inside the return.
    :return:
    """
    data = {"sub_state_run": [{"arbitrary": "data", "changes": {}}]}
    out = statedecorators.OutputUnifier("content_check")(lambda: data)()
    assert (
        " The following keys were not present in the state return: name, result, comment"
        in out["sub_state_run"][0]["comment"]
    )
    assert not out["sub_state_run"][0]["result"]


def test_sub_state_output_unifier_comment_is_not_list():
    """
    Test for output is unified so the comment is converted to a multi-line string
    :return:
    """
    data = {
        "sub_state_run": [
            {
                "comment": ["data", "in", "the", "list"],
                "changes": {},
                "name": None,
                "result": "fantastic!",
            }
        ]
    }
    expected = {
        "sub_state_run": [
            {
                "comment": "data\nin\nthe\nlist",
                "changes": {},
                "name": None,
                "result": True,
            }
        ]
    }
    assert statedecorators.OutputUnifier("unify")(lambda: data)() == expected

    data = {
        "sub_state_run": [
            {
                "comment": ["data", "in", "the", "list"],
                "changes": {},
                "name": None,
                "result": None,
            }
        ]
    }
    expected = "data\nin\nthe\nlist"
    assert (
        statedecorators.OutputUnifier("unify")(lambda: data)()["sub_state_run"][0][
            "comment"
        ]
        == expected
    )


def test_sub_state_output_unifier_result_converted_to_true():
    """
    Test for output is unified so the result is converted to True
    :return:
    """
    data = {
        "sub_state_run": [
            {
                "comment": ["data", "in", "the", "list"],
                "changes": {},
                "name": None,
                "result": "Fantastic",
            }
        ]
    }
    assert (
        statedecorators.OutputUnifier("unify")(lambda: data)()["sub_state_run"][0][
            "result"
        ]
        is True
    )


def test_sub_state_output_unifier_result_converted_to_false():
    """
    Test for output is unified so the result is converted to False
    :return:
    """
    data = {
        "sub_state_run": [
            {
                "comment": ["data", "in", "the", "list"],
                "changes": {},
                "name": None,
                "result": "",
            }
        ]
    }
    assert (
        statedecorators.OutputUnifier("unify")(lambda: data)()["sub_state_run"][0][
            "result"
        ]
        is False
    )



"""
Retrieve Pillar data by running a SQLCipher query

.. versionadded:: 2016.3.0

Python SQLCipher support is provided by the pysqlcipher
Python package. You need this module installed to query
Pillar data from a SQLCipher database.

This module is a concrete implementation of the sql_base
ext_pillar for SQLCipher.

:maturity: new
:depends: pysqlcipher (for py2) or pysqlcipher3 (for py3)
:platform: all

Configuring the sqlcipher ext_pillar
====================================

Use the 'sqlcipher' key under ext_pillar for configuration of queries.

SQLCipher database connection configuration requires the following values
configured in the master config:

   * ``sqlcipher.database`` - The SQLCipher database to connect to.
     Defaults to ``'/var/lib/salt/pillar-sqlcipher.db'``.
   * ``sqlcipher.pass`` - The SQLCipher database decryption password.
   * ``sqlcipher.timeout`` - The connection timeout in seconds.

Example configuration

.. code-block:: yaml

    sqlcipher:
      database: /var/lib/salt/pillar-sqlcipher.db
      pass: strong_pass_phrase
      timeout: 5.0

Complete Example
================

.. code-block:: yaml

    sqlcipher:
      database: '/var/lib/salt/pillar-sqlcipher.db'
      pass: strong_pass_phrase
      timeout: 5.0

    ext_pillar:
      - sqlcipher:
          fromdb:
            query: 'SELECT col1,col2,col3,col4,col5,col6,col7
                      FROM some_random_table
                     WHERE minion_pattern LIKE ?'
            depth: 5
            as_list: True
            with_lists: [1,3]
"""

import logging
from contextlib import contextmanager

from salt.pillar.sql_base import SqlBaseExtPillar

# Set up logging
log = logging.getLogger(__name__)

try:
    from pysqlcipher import dbapi2 as sqlcipher

    HAS_SQLCIPHER = True
except ImportError:
    HAS_SQLCIPHER = False


def __virtual__():
    if not HAS_SQLCIPHER:
        return False
    return True


class SQLCipherExtPillar(SqlBaseExtPillar):
    """
    This class receives and processes the database rows from SQLCipher.
    """

    @classmethod
    def _db_name(cls):
        return "SQLCipher"

    def _get_options(self):
        """
        Returns options used for the SQLCipher connection.
        """
        defaults = {
            "database": "/var/lib/salt/pillar-sqlcipher.db",
            "pass": "strong_pass_phrase",
            "timeout": 5.0,
        }
        _options = {}
        _opts = __opts__.get("sqlcipher", {})

        for attr in defaults:
            if attr not in _opts:
                log.debug("Using default for SQLCipher pillar %s", attr)
                _options[attr] = defaults[attr]
                continue
            _options[attr] = _opts[attr]
        return _options

    @contextmanager
    def _get_cursor(self):
        """
        Yield a SQLCipher cursor
        """
        _options = self._get_options()
        conn = sqlcipher.connect(
            _options.get("database"), timeout=float(_options.get("timeout"))
        )
        conn.execute('pragma key="{}"'.format(_options.get("pass")))
        cursor = conn.cursor()
        try:
            yield cursor
        except sqlcipher.Error as err:
            log.exception("Error in ext_pillar SQLCipher: %s", err.args)
        finally:
            conn.close()


def ext_pillar(minion_id, pillar, *args, **kwargs):
    """
    Execute queries against SQLCipher, merge and return as a dict
    """
    return SQLCipherExtPillar().fetch(minion_id, pillar, *args, **kwargs)



"""
Unit tests for the Default Job Cache (local_cache).
"""

import logging
import os
import time

import pytest

import salt.returners.local_cache as local_cache
import salt.utils.files
import salt.utils.jid
import salt.utils.job
import salt.utils.platform
from tests.support.mock import patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(tmp_cache_dir):
    return {
        local_cache: {
            "__opts__": {
                "cachedir": str(tmp_cache_dir),
                "keep_jobs_seconds": 0.0000000010,
            }
        }
    }


@pytest.fixture
def tmp_cache_dir(tmp_path):
    return tmp_path / "cache_dir"


@pytest.fixture
def jobs_dir(tmp_cache_dir):
    return tmp_cache_dir / "jobs"


@pytest.fixture
def jid_dir(jobs_dir):
    return (
        jobs_dir
        / "31"
        / "c56eed380a4e899ae12bc42563cfdfc53066fb4a6b53e2378a08ac49064539"
    )


@pytest.fixture
def pki_dir(tmp_path):
    dirname = tmp_path / "pki_dir"
    dirname.mkdir(parents=True, exist_ok=True)
    (dirname / "minion").touch()
    return dirname


@pytest.fixture
def job_cache_dir_files(jid_dir):
    return [str(jid_dir / "jid"), str(jid_dir / "minion" / "return.p")]


def _check_dir_files(msg, contents, status="None"):
    """
    helper method to ensure files or dirs
    are either present or removed
    """
    for content in contents:
        log.debug("CONTENT %s", content)
        if status == "present":
            check_job_dir = os.path.exists(content)
        elif status == "removed":
            if os.path.exists(content):
                check_job_dir = False
            else:
                check_job_dir = True
        assert check_job_dir, msg + content


@pytest.fixture
def add_job(tmp_cache_dir, job_cache_dir_files, pki_dir, tmp_path):
    def _add_job():
        """
        helper method to add job.
        """
        # add the job.
        opts = {
            "cachedir": str(tmp_cache_dir),
            "master_job_cache": "local_cache",
            "pki_dir": str(pki_dir),
            "conf_file": str(tmp_path / "conf"),
            "job_cache": True,
        }
        load = {
            "fun_args": [],
            "jid": "20160603132323715452",
            "return": True,
            "retcode": 0,
            "success": True,
            "cmd": "_return",
            "fun": "test.ping",
            "id": "minion",
        }

        add_job = salt.utils.job.store_job(opts, load)
        assert add_job is None
        _check_dir_files(
            "Dir/file does not exist: ", job_cache_dir_files, status="present"
        )

    return _add_job


@pytest.mark.slow_test
def test_clean_old_jobs(add_job, job_cache_dir_files):
    """
    test to ensure jobs are removed from job cache
    """
    add_job()

    if salt.utils.platform.is_windows():
        time.sleep(0.01)

    # remove job
    assert local_cache.clean_old_jobs() is None

    _check_dir_files(
        "job cache was not removed: ", job_cache_dir_files, status="removed"
    )


@pytest.mark.slow_test
def test_not_clean_new_jobs(add_job, job_cache_dir_files):
    """
    test to ensure jobs are not removed when
    jobs dir is new
    """
    add_job()

    with patch.dict(local_cache.__opts__, {"keep_jobs_seconds": 86400}):
        assert local_cache.clean_old_jobs() is None

        _check_dir_files(
            "job cache was removed: ", job_cache_dir_files, status="present"
        )


@pytest.mark.slow_test
def test_override_clean_jobs(add_job, job_cache_dir_files):
    """
    test to ensure keep_jobs_seconds overrides keep_jobs if set
    """
    add_job()
    time.sleep(1.5)

    with patch.dict(local_cache.__opts__, {"keep_jobs_seconds": 1, "keep_jobs": 4}):
        assert local_cache.clean_old_jobs() is None

        _check_dir_files(
            "job cache was removed: ", job_cache_dir_files, status="removed"
        )


@pytest.mark.slow_test
def test_override_clean_jobs_seconds(add_job, job_cache_dir_files):
    """
    test to ensure keep_jobs still works as long as keep_jobs_seconds is set to default
    """
    add_job()

    with patch.dict(local_cache.__opts__, {"keep_jobs_seconds": 86400, "keep_jobs": 1}):
        assert local_cache.clean_old_jobs() is None

        _check_dir_files(
            "job cache was not removed: ", job_cache_dir_files, status="present"
        )


@pytest.mark.slow_test
def test_empty_jid_dir(jobs_dir):
    """
    test to ensure removal of empty jid dir
    """
    # add empty jid dir
    empty_jid_dir = []
    new_jid_dir = jobs_dir / "z0"
    new_jid_dir.mkdir(parents=True, exist_ok=True)
    new_jid_dir = str(new_jid_dir)
    empty_jid_dir.append(new_jid_dir)

    # This needed due to a race condition in Windows
    # `os.makedirs` hasn't released the handle before
    # `local_cache.clean_old_jobs` tries to delete the new_jid_dir
    if salt.utils.platform.is_windows():
        import time

        lock_dir = new_jid_dir + ".lckchk"
        tries = 0
        while True:
            tries += 1
            if tries > 10:
                break
            # Rename the directory and name it back
            # If it fails, the directory handle is not released, try again
            # If it succeeds, break and continue test
            try:
                os.rename(new_jid_dir, lock_dir)
                time.sleep(1)
                os.rename(lock_dir, new_jid_dir)
                break
            except OSError:  # pylint: disable=E0602
                continue

    # check dir exists
    _check_dir_files("new_jid_dir was not created", empty_jid_dir, status="present")

    # remove job
    assert local_cache.clean_old_jobs() is None

    # check jid dir is removed
    _check_dir_files("new_jid_dir was not removed", empty_jid_dir, status="removed")
